"""The unpolarized, time-like |NLO| Altarelli-Parisi splitting kernels."""

import numba as nb
import numpy as np
from numpy import power as npp

from eko import constants
from eko.constants import zeta2, zeta3

from ....harmonics import cache as c


@nb.njit(cache=True)
def gamma_nsp(N, nf, cache):
    r"""Compute the |NLO| non-singlet positive anomalous dimension.

    Implements :eqref:`A6` from :cite:`Gluck:1992zx`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        NLO non-singlet positive anomalous dimension
        :math:`\gamma_{ns}^{(1)+}(N)`

    """
    NS = N * N
    NT = NS * N
    NFO = NT * N

    N1 = N + 1
    N1S = N1 * N1
    N1T = N1S * N1

    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    S1h = c.get(c.S1h, cache, N)
    S2h = c.get(c.S2h, cache, N)
    S3h = c.get(c.S3h, cache, N)
    S1ph = c.get(c.S1ph, cache, N)
    g3 = c.get(c.g3, cache, N)

    SLC = -5 / 8 * zeta3
    SLV = g3 + S1 / N**2 - (zeta2 / 2) * (-2 / (1 + N) + 2 / N + S1ph - S1h)

    SSCHLP = SLC + SLV

    PNPA = (
        16 * S1 * (2 * N + 1) / (NS * N1S)
        + 16 * (2 * S1 - 1 / (N * N1)) * (S2 - S2h)
        + 64 * SSCHLP
        + 24 * S2
        - 3
        - 8 * S3h
        - 8 * (3 * NT + NS - 1) / (NT * N1T)
        - 16 * (2 * NS + 2 * N + 1) / (NT * N1T)
    ) * (-0.5)
    PNSB = (
        S1 * (536 / 9 + 8 * (2 * N + 1) / (NS * N1S))
        - (16 * S1 + 52 / 3 - 8 / (N * N1)) * S2
        - 43 / 6
        - (151 * NFO + 263 * NT + 97 * NS + 3 * N + 9) * 4 / (9 * NT * N1T)
    ) * (-0.5)
    PNSC = (
        -160 / 9 * S1
        + 32 / 3.0 * S2
        + 4 / 3
        + 16 * (11 * NS + 5 * N - 3) / (9 * NS * N1S)
    ) * (-0.5)
    PNSTL = (-4 * S1 + 3 + 2 / (N * N1)) * (
        2 * S2 - 2 * zeta2 - (2 * N + 1) / (NS * N1S)
    )

    result = (
        constants.CF
        * (
            (constants.CF - constants.CA / 2) * PNPA
            + constants.CA * PNSB
            + (1 / 2) * nf * PNSC
        )
        + constants.CF**2 * PNSTL * 4
    )
    return -result


@nb.njit(cache=True)
def gamma_nsm(N, nf, cache):
    r"""Compute the |NLO| non-singlet negative anomalous dimension.

    Based on :cite:`Gluck:1992zx`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        NLO non-singlet negative anomalous dimension
        :math:`\gamma_{ns}^{(1)-}(N)`

    """
    NS = N * N
    NT = NS * N
    NFO = NT * N

    N1 = N + 1
    N1S = N1 * N1
    N1T = N1S * N1

    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    S1h = c.get(c.S1h, cache, N)
    S2mh = c.get(c.S2mh, cache, N)
    S3mh = c.get(c.S3mh, cache, N)
    S1ph = c.get(c.S1ph, cache, N)
    g3 = c.get(c.g3, cache, N)

    SLC = -5 / 8 * zeta3
    SLV = g3 + S1 / N**2 - (zeta2 / 2) * (-2 / (1 + N) + 2 / N + S1ph - S1h)
    SSCHLM = SLC - SLV

    PNMA = (
        16 * S1 * (2 * N + 1) / (NS * N1S)
        + 16 * (2 * S1 - 1 / (N * N1)) * (S2 - S2mh)
        + 64 * SSCHLM
        + 24 * S2
        - 3
        - 8 * S3mh
        - 8 * (3 * NT + NS - 1) / (NT * N1T)
        + 16 * (2 * NS + 2 * N + 1) / (NT * N1T)
    ) * (-0.5)
    PNSB = (
        S1 * (536 / 9 + 8 * (2 * N + 1) / (NS * N1S))
        - (16 * S1 + 52 / 3 - 8 / (N * N1)) * S2
        - 43 / 6
        - (151 * NFO + 263 * NT + 97 * NS + 3 * N + 9) * 4 / (9 * NT * N1T)
    ) * (-0.5)
    PNSC = (
        -160 / 9 * S1
        + 32 / 3.0 * S2
        + 4 / 3
        + 16 * (11 * NS + 5 * N - 3) / (9 * NS * N1S)
    ) * (-0.5)
    PNSTL = (-4 * S1 + 3 + 2 / (N * N1)) * (
        2 * S2 - 2 * zeta2 - (2 * N + 1) / (NS * N1S)
    )

    result = (
        constants.CF
        * (
            (constants.CF - constants.CA / 2) * PNMA
            + constants.CA * PNSB
            + (1 / 2) * nf * PNSC
        )
        + constants.CF**2 * PNSTL * 4
    )
    return -result


@nb.njit(cache=True)
def gamma_qqs(N, nf):
    r"""Compute the |NLO| quark-quark singlet anomalous dimension.

    Implements :eqref:`B.9` from :cite:`Mitov:2006wy`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors

    Returns
    -------
    complex
        NLO quark-quark singlet anomalous dimension
        :math:`\gamma_{qq}^{(1)s}(N)`

    """
    qqS1 = (
        constants.CF
        * nf
        * (
            (
                4
                * (
                    8
                    + 44 * N
                    + 46 * npp(N, 2)
                    + 21 * npp(N, 3)
                    + 14 * npp(N, 4)
                    + 15 * npp(N, 5)
                    + 10 * npp(N, 6)
                    + 2 * npp(N, 7)
                )
            )
            / ((-1 + N) * npp(N, 3) * npp(1 + N, 3) * npp(2 + N, 2))
        )
    )
    return qqS1


@nb.njit(cache=True)
def gamma_qg(N, nf, cache):
    r"""Compute the |NLO| quark-gluon anomalous dimension.

    Implements :eqref:`B.10` from :cite:`Mitov:2006wy`
    and :eqref:`A1` from :cite:`Gluck:1992zx`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        NLO quark-gluon anomalous dimension
        :math:`\gamma_{qg}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    Sm2 = c.get(c.Sm2, cache, N, is_singlet=True)

    qg1 = (
        nf
        * constants.CF
        * (
            (
                2
                * (
                    8
                    + 12 * N
                    + 18 * npp(N, 2)
                    + 77 * npp(N, 3)
                    + 127 * npp(N, 4)
                    + 104 * npp(N, 5)
                    + 45 * npp(N, 6)
                    + 9 * npp(N, 7)
                )
            )
            / (npp(N, 3) * npp(1 + N, 3) * npp(2 + N, 2))
            - (
                4
                * S1
                * (
                    -8
                    - 12 * N
                    + 22 * npp(N, 2)
                    + 25 * npp(N, 3)
                    + 10 * npp(N, 4)
                    + 3 * npp(N, 5)
                )
            )
            / (npp(N, 2) * npp(1 + N, 2) * npp(2 + N, 2))
            + (4 * npp(S1, 2) * (2 + N + npp(N, 2))) / (N * (1 + N) * (2 + N))
            - (20 * S2 * (2 + N + npp(N, 2))) / (N * (1 + N) * (2 + N))
        )
    )
    qg2 = (
        nf
        * constants.CA
        * (
            (
                4
                * (
                    144
                    + 600 * N
                    + 980 * npp(N, 2)
                    + 2366 * npp(N, 3)
                    + 2564 * npp(N, 4)
                    + 379 * npp(N, 5)
                    - 1177 * npp(N, 6)
                    - 1037 * npp(N, 7)
                    - 423 * npp(N, 8)
                    - 76 * npp(N, 9)
                    - 288 * zeta2 * npp(N, 2)
                    - 720 * zeta2 * npp(N, 3)
                    - 504 * zeta2 * npp(N, 4)
                    + 180 * zeta2 * npp(N, 5)
                    + 576 * zeta2 * npp(N, 6)
                    + 504 * zeta2 * npp(N, 7)
                    + 216 * zeta2 * npp(N, 8)
                    + 36 * zeta2 * npp(N, 9)
                    - 180 * npp(N, 3)
                    - 72 * npp(N, 4)
                    + 108 * npp(N, 5)
                    + 108 * npp(N, 6)
                    + 36 * npp(N, 7)
                )
            )
            / (9 * (-1 + N) * npp(N, 3) * npp(1 + N, 3) * npp(2 + N, 3))
            + (8 * Sm2 * (2 + N + npp(N, 2))) / (N * (1 + N) * (2 + N))
            + (
                4
                * S1
                * (
                    -48
                    - 100 * N
                    + 40 * npp(N, 2)
                    + 77 * npp(N, 3)
                    + 32 * npp(N, 4)
                    + 11 * npp(N, 5)
                )
            )
            / (3 * npp(N, 2) * npp(1 + N, 2) * npp(2 + N, 2))
            - (4 * npp(S1, 2) * (2 + N + npp(N, 2))) / (N * (1 + N) * (2 + N))
            + (12 * S2 * (2 + N + npp(N, 2))) / (N * (1 + N) * (2 + N))
        )
    )
    qg3 = (
        nf
        * nf
        * (
            (
                8
                * (
                    -12
                    - 16 * N
                    + 37 * npp(N, 2)
                    + 41 * npp(N, 3)
                    + 17 * npp(N, 4)
                    + 5 * npp(N, 5)
                )
            )
            / (9 * npp(N, 2) * npp(1 + N, 2) * npp(2 + N, 2))
            - (8 * S1 * (2 + N + npp(N, 2))) / (3 * N * (1 + N) * (2 + N))
        )
    )
    result = (1 / (2 * nf)) * (qg1 + qg2 + qg3)
    return result


@nb.njit(cache=True)
def gamma_gq(N, nf, cache):
    r"""Compute the |NLO| gluon-quark anomalous dimension.

    Implements :eqref:`B.11` from :cite:`Mitov:2006wy`
    and :eqref:`A1` from :cite:`Gluck:1992zx`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        NLO gluon-quark anomalous dimension
        :math:`\gamma_{gq}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    Sm2 = c.get(c.Sm2, cache, N, is_singlet=True)
    gq1 = (
        constants.CF
        * constants.CF
        * (
            (
                2
                * (
                    -4
                    - 4 * N
                    + 41 * npp(N, 2)
                    + 83 * npp(N, 3)
                    + 41 * npp(N, 4)
                    - 11 * npp(N, 5)
                    - 10 * npp(N, 6)
                    - 8 * npp(N, 7)
                    - 16 * zeta2 * npp(N, 2)
                    - 24 * zeta2 * npp(N, 3)
                    + 16 * zeta2 * npp(N, 5)
                    + 16 * zeta2 * npp(N, 6)
                    + 8 * zeta2 * npp(N, 7)
                )
            )
            / (npp(-1 + N, 2) * npp(N, 3) * npp(1 + N, 3))
            + (
                8
                * (4 - 2 * N - 16 * npp(N, 2) - npp(N, 3) - 2 * npp(N, 4) + npp(N, 5))
                * S1
            )
            / (npp(-1 + N, 2) * npp(N, 2) * npp(1 + N, 2))
            - (4 * (2 + N + npp(N, 2)) * npp(S1, 2)) / ((-1 + N) * N * (1 + N))
            + (12 * (2 + N + npp(N, 2)) * S2) / ((-1 + N) * N * (1 + N))
        )
    )
    gq2 = (
        constants.CF
        * constants.CA
        * (
            -(
                4
                * (
                    16
                    - 144 * npp(N, 2)
                    - 156 * npp(N, 3)
                    - 101 * npp(N, 4)
                    - 77 * npp(N, 5)
                    - 75 * npp(N, 6)
                    - 44 * npp(N, 7)
                    - npp(N, 8)
                    + 5 * npp(N, 9)
                    + npp(N, 10)
                    + 16 * N
                    + 32 * npp(N, 2)
                    - 20 * npp(N, 3)
                    - 44 * npp(N, 4)
                    - 26 * npp(N, 5)
                    + 14 * npp(N, 6)
                    + 22 * npp(N, 7)
                    + 6 * npp(N, 8)
                )
            )
            / (npp(-1 + N, 3) * npp(N, 3) * npp(1 + N, 3) * npp(2 + N, 2))
            + (8 * (2 + N + npp(N, 2)) * Sm2) / ((-1 + N) * N * (1 + N))
            - (8 * (2 - 2 * N - 9 * npp(N, 2) + npp(N, 3) - npp(N, 4) + npp(N, 5)) * S1)
            / (npp(-1 + N, 2) * npp(N, 2) * npp(1 + N, 2))
            + (4 * (2 + N + npp(N, 2)) * npp(S1, 2)) / ((-1 + N) * N * (1 + N))
            - (20 * (2 + N + npp(N, 2)) * S2) / ((-1 + N) * N * (1 + N))
        )
    )
    result = (2 * nf) * (gq1 + gq2)
    return result


@nb.njit(cache=True)
def gamma_gg(N, nf, cache):
    r"""Compute the |NLO| gluon-gluon anomalous dimension.

    Implements :eqref:`B.12` from :cite:`Mitov:2006wy`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    complex
        NLO gluon-gluon anomalous dimension
        :math:`\gamma_{gg}^{(1)}(N)`

    """
    S1 = c.get(c.S1, cache, N)
    S2 = c.get(c.S2, cache, N)
    S3 = c.get(c.S3, cache, N)
    Sm2 = c.get(c.Sm2, cache, N, is_singlet=True)
    Sm3 = c.get(c.Sm3, cache, N, is_singlet=True)
    Sm21 = c.get(c.Sm21, cache, N, is_singlet=True)

    gg1 = (
        nf
        * constants.CF
        * (
            (
                2
                * (
                    -16
                    + 8 * N
                    + 108 * npp(N, 2)
                    + 162 * npp(N, 3)
                    + 106 * npp(N, 4)
                    + 11 * npp(N, 5)
                    - 5 * npp(N, 6)
                    - 2 * npp(N, 7)
                    + 6 * npp(N, 8)
                    + 5 * npp(N, 9)
                    + npp(N, 10)
                )
            )
            / (npp(-1 + N, 2) * npp(N, 3) * npp(1 + N, 3) * npp(2 + N, 2))
        )
    )
    gg2 = (
        constants.CA
        * constants.CA
        * (
            (
                2
                * (
                    -576
                    + 240 * N
                    + 3824 * npp(N, 2)
                    + 1240 * npp(N, 3)
                    + 1928 * npp(N, 4)
                    + 8303 * npp(N, 5)
                    + 10651 * npp(N, 6)
                    + 6614 * npp(N, 7)
                    + 1238 * npp(N, 8)
                    - 1133 * npp(N, 9)
                    - 889 * npp(N, 10)
                    - 288 * npp(N, 11)
                    - 48 * npp(N, 12)
                    + 1152 * zeta2 * npp(N, 2)
                    + 1248 * zeta2 * npp(N, 3)
                    - 1296 * zeta2 * npp(N, 4)
                    - 792 * zeta2 * npp(N, 5)
                    + 876 * zeta2 * npp(N, 6)
                    - 1368 * zeta2 * npp(N, 7)
                    - 2340 * zeta2 * npp(N, 8)
                    + 120 * zeta2 * npp(N, 9)
                    + 1476 * zeta2 * npp(N, 10)
                    + 792 * zeta2 * npp(N, 11)
                    + 132 * zeta2 * npp(N, 12)
                    - 576 * N
                    - 1440 * npp(N, 2)
                    + 216 * npp(N, 3)
                    + 1800 * npp(N, 4)
                    + 1800 * npp(N, 5)
                    - 72 * npp(N, 6)
                    - 1008 * npp(N, 7)
                    - 576 * npp(N, 8)
                    - 144 * npp(N, 9)
                )
            )
            / (9 * npp(-1 + N, 3) * npp(N, 3) * npp(1 + N, 3) * npp(2 + N, 3))
            - (8 * Sm3)
            + (Sm2)
            * (
                (32 * (1 + N + npp(N, 2))) / ((-1 + N) * N * (1 + N) * (2 + N))
                - 16 * S1
            )
            - (8 * S2 * (12 - 10 * N + npp(N, 2) + 22 * npp(N, 3) + 11 * npp(N, 4)))
            / (3 * (-1 + N) * N * (1 + N) * (2 + N))
            + (S1)
            * (
                -(
                    4
                    * (
                        -144
                        - 144 * N
                        + 236 * npp(N, 2)
                        + 308 * npp(N, 3)
                        + 829 * npp(N, 4)
                        + 680 * npp(N, 5)
                        - 134 * npp(N, 6)
                        - 268 * npp(N, 7)
                        - 67 * npp(N, 8)
                        + 288 * zeta2 * npp(N, 2)
                        + 288 * zeta2 * npp(N, 3)
                        - 504 * zeta2 * npp(N, 4)
                        - 576 * zeta2 * npp(N, 5)
                        + 144 * zeta2 * npp(N, 6)
                        + 288 * zeta2 * npp(N, 7)
                        + 72 * zeta2 * npp(N, 8)
                    )
                )
                / (9 * npp(-1 + N, 2) * npp(N, 2) * npp(1 + N, 2) * npp(2 + N, 2))
                + 16 * S2
            )
            - (8 * S3)
            + (16 * Sm21)
        )
    )
    gg3 = (
        nf
        * constants.CA
        * (
            -(
                8
                * (
                    -12
                    + 26 * N
                    + 132 * npp(N, 2)
                    + 85 * npp(N, 3)
                    + 34 * npp(N, 4)
                    - 9 * npp(N, 5)
                    - 25 * npp(N, 6)
                    - 12 * npp(N, 7)
                    - 3 * npp(N, 8)
                    + 24 * zeta2 * npp(N, 2)
                    + 24 * zeta2 * npp(N, 3)
                    - 42 * zeta2 * npp(N, 4)
                    - 48 * zeta2 * npp(N, 5)
                    + 12 * zeta2 * npp(N, 6)
                    + 24 * zeta2 * npp(N, 7)
                    + 6 * zeta2 * npp(N, 8)
                )
            )
            / (9 * npp(-1 + N, 2) * npp(N, 2) * npp(1 + N, 2) * npp(2 + N, 2))
            - (S1 * (40 / 9))
            + (S2 * (16 / 3))
        )
    )
    result = gg1 + gg2 + gg3
    return result


@nb.njit(cache=True)
def gamma_singlet(N, nf, cache):
    r"""Compute the |NLO| singlet anomalous dimension matrix.

    Implements :eqref:`2.13` from :cite:`Gluck:1992zx`.

    Parameters
    ----------
    N : complex
        Mellin moment
    nf : int
        No. of active flavors
    cache: numpy.ndarray
        Harmonic sum cache

    Returns
    -------
    numpy.ndarray
        NLO singlet anomalous dimension matrix
        :math:`\gamma_{s}^{(1)}`

    """
    gamma_qq = gamma_nsp(N, nf, cache) + gamma_qqs(N, nf)

    result = np.array(
        [
            [gamma_qq, gamma_gq(N, nf, cache)],
            [gamma_qg(N, nf, cache), gamma_gg(N, nf, cache)],
        ],
        np.complex_,
    )
    return result



r"""Contains the |NNLO| |OME| in the polarized case for the matching conditions in the |VFNS|.

The equations are given in :cite:`Bierenbaum:2022biv`.
As in the |NLO| |OME|, in the paper, an additional factor 2 can be found in front of the anomalous dimensions and factor (-1) for odd powers of L.
The anomalous dimensions and beta function with the addition 'hat' are defined as in the |NLO| case.
"""
import numba as nb
import numpy as np

from eko.constants import CA, CF, TR, zeta2, zeta3

from ....anomalous_dimensions.polarized.space_like.as1 import gamma_gq as gamma0_gq
from ....anomalous_dimensions.polarized.space_like.as1 import gamma_ns as gamma0_qq
from ....anomalous_dimensions.polarized.space_like.as1 import gamma_qg as gamma0_qg
from ....anomalous_dimensions.polarized.space_like.as2 import gamma_qg as gamma1_qg
from ....harmonics import cache as c

beta_0hat = -4 / 3 * TR
"""This is the lowest order beta function with the addition 'hat' defined as above."""


@nb.njit(cache=True)
def A_qq_ns(n, cache, L):
    r"""Compute |NNLO| light-light non-singlet |OME| :math:`A_{qq,H}^{NS,(2)}`.

    Implements :eqref:`133` of :cite:`Bierenbaum:2022biv`.

    Parameters
    ----------
    n : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    complex
        |NNLO| light-light non-singlet |OME| :math:`A_{qq,H}^{NS,(2)}`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    S3 = c.get(c.S3, cache, n)

    gamma1_qqNS = (
        -CF
        * TR
        * (4 / 3)
        * (
            -8 * S2
            + (40 / 3) * S1
            - (3 * n**4 + 6 * n**3 + 47 * n**2 + 20 * n - 12)
            / (3 * n**2 * (n + 1) ** 2)
        )
    )
    aqqns = (
        TR
        * CF
        * (
            -(8 / 3) * S3
            - (8 / 3) * zeta2 * S1
            + (40 / 9) * S2
            + 2 * zeta2 * ((3 * n**2 + 3 * n + 2) / (3 * n * (n + 1)))
            - (224 * S1 / 27)
            + (
                219 * n**6
                + 657 * n**5
                + 1193 * n**4
                + 763 * n**3
                - 40 * n**2
                - 48 * n
                + 72
            )
            / (54 * n**3 * (n + 1) ** 3)
        )
    )

    a_qq_l0 = aqqns - 1 / 4 * 2 * beta_0hat * gamma0_qq(n, cache) * zeta2
    a_qq_l1 = (1 / 2) * gamma1_qqNS
    a_qq_l2 = (1 / 4) * beta_0hat * 2 * gamma0_qq(n, cache)
    return a_qq_l2 * L**2 + a_qq_l1 * (-L) + a_qq_l0


@nb.njit(cache=True)
def A_hq_ps(n, cache, L, nf):
    r"""Compute |NNLO| heavy-light pure-singlet |OME| :math:`A_{Hq}^{PS,(2)}`.

    Implements :eqref:`138` of :cite:`Bierenbaum:2022biv`.

    Parameters
    ----------
    n : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`
    nf : int
        Number of active flavors

    Returns
    -------
    complex
        |NNLO| heavy-light pure-singlet |OME| :math:`A_{Hq}^{PS,(2)}`

    """
    S2 = c.get(c.S2, cache, n)

    a_hq_ps_l = (
        -4
        * TR
        * CF
        * ((n + 2) / (n**2 * (n + 1) ** 2))
        * (
            (n - 1) * (2 * S2 + zeta2)
            - ((4 * n**3 - 4 * n**2 - 3 * n - 1) / (n**2 * (n + 1) ** 2))
        )
    )
    # term that differentiates between M scheme and Larin scheme,
    # we are computing in M scheme hence the addition of this term
    z_qq_ps = (
        -CF * TR * nf * ((8 * (2 + n) * (n**2 - n - 1)) / (n**3 * (n + 1) ** 3))
    )
    gamma1_ps_qqhat = (16 * CF * (2 + n) * (1 + 2 * n + n**3) * TR) / (
        n**3 * ((1 + n) ** 3)
    )
    a_hq_l0 = (
        a_hq_ps_l
        + z_qq_ps
        + (zeta2 / 8) * (2 * gamma0_qg(n, nf=1)) * (2 * gamma0_gq(n))
    )
    a_hq_l1 = (1 / 2) * gamma1_ps_qqhat
    a_hq_l2 = -(1 / 8) * (2 * gamma0_qg(n, nf=1)) * (2 * gamma0_gq(n))
    return a_hq_l2 * L**2 + a_hq_l1 * (-L) + a_hq_l0


@nb.njit(cache=True)
def A_hg(n, cache, L):
    r"""Compute |NNLO| heavy-gluon |OME| :math:`A_{Hg}^{S,(2)}`.

    Implements :eqref:`111` of :cite:`Bierenbaum:2022biv`.

    Parameters
    ----------
    n : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    complex
        |NNLO| heavy-gluon |OME| :math:`A_{Hg}^{S,(2)}`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)
    S3 = c.get(c.S3, cache, n)
    Sm21 = c.get(c.Sm21, cache, n, is_singlet=False)
    S2h = c.get(c.S2h, cache, n)
    S3h = c.get(c.S3h, cache, n)
    S2ph = c.get(c.S2ph, cache, n)
    S3ph = c.get(c.S3ph, cache, n)

    a_hg = (
        (1 / (6 * n**4 * (1 + n) ** 4 * (2 + n)))
        * TR
        * (
            -2
            * CF
            * (
                3
                * (
                    4
                    + 8 * n
                    + 3 * n**2
                    - 2 * n**3
                    - 25 * n**4
                    + 60 * n**6
                    + 52 * n**7
                    + 12 * n**8
                )
                + (-1 + n)
                * n**2
                * (1 + n) ** 2
                * (2 + n)
                * (2 + 3 * n + 3 * n**3)
                * np.pi**2
                + 12 * n**2 * (1 + n) ** 3 * (-36 - 22 * n - 2 * n**2 + n**3) * S1
                + 12 * n**2 * (1 + n) ** 3 * (-2 + 3 * n + 3 * n**2) * S1**2
                + 12
                * n**2
                * (1 + n) ** 2
                * (2 + 33 * n + 43 * n**2 + 17 * n**3 + n**4)
                * S2
                - 4
                * (-1 + n)
                * n**3
                * (1 + n) ** 3
                * (2 + n)
                * (S1**3 + S1 * (np.pi**2 + 3 * S2) - 4 * S3)
            )
            + CA
            * (
                -24
                * (1 + n) ** 3
                * (
                    4
                    + 12 * n
                    + 33 * n**2
                    + 4 * n**3
                    + 29 * n**4
                    + 36 * n**5
                    + 22 * n**6
                    + 10 * n**7
                    + 2 * n**8
                )
                + 8 * (-1 + n) * n**2 * (1 + n) ** 2 * (2 + n) ** 2 * np.pi**2
                + 24
                * n**3
                * (1 + n)
                * (2 - 10 * n - n**2 + 4 * n**3 + n**4)
                * S1
                + 24 * n**3 * (1 + n) ** 2 * (5 + 4 * n + n**2) * S1**2
                + 24
                * n**2
                * (1 + n) ** 2
                * (-16 + 15 * n + 24 * n**2 + 7 * n**3)
                * S2
                - 24
                * (1 - n)
                * n**3
                * (2 + n)
                * (4 + (1 + n) ** 2 * S2h - (1 + n) ** 2 * S2ph)
                + (1 - n)
                * n**3
                * (2 + n)
                * (
                    -48
                    + 8 * (1 + n) ** 3 * S1**3
                    + 72 * (1 + n) ** 3 * S1 * S2
                    - 24 * (1 + n) * S1 * (4 + (1 + n) ** 2 * S2h - (1 + n) ** 2 * S2ph)
                    + 64 * (1 + n) ** 3 * S3
                    + 6 * (1 + n) ** 3 * S3ph
                    - 6 * (1 + n) ** 3 * (S3h - zeta3)
                    + 18 * (1 + n) ** 3 * zeta3
                    - 12 * (1 + n) ** 3 * (8 * Sm21 + 5 * zeta3)
                )
            )
        )
    )

    # remove the nf dependence from
    # (2 * gamma0_gg(n, S1, nf) + 2 * beta_0(nf))
    pgg0_beta0 = 8 * CA * (-(2 / (n + n**2)) + S1)
    a_hg_l0 = a_hg + (1 / 8) * 2 * gamma0_qg(n, nf=1) * (
        pgg0_beta0 - 2 * gamma0_qq(n, cache)
    )
    a_hg_l1 = 2 * gamma1_qg(n, nf=1, cache=cache)
    a_hg_l2 = (
        (1 / 8)
        * 2
        * gamma0_qg(n, nf=1)
        * (2 * gamma0_qq(n, cache) - pgg0_beta0 - 4 * beta_0hat)
    )
    return a_hg_l2 * L**2 + a_hg_l1 * (-L) + a_hg_l0


@nb.njit(cache=True)
def A_gq(n, cache, L):
    r"""Compute |NNLO| gluon-quark |OME| :math:`A_{gq,H}^{S,(2)}`.

    Implements :eqref:`174` of :cite:`Bierenbaum:2022biv`.

    Parameters
    ----------
    n : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    complex
        |NNLO| gluon-quark |OME| :math:`A_{gq,H}^{S,(2)}`

    """
    S1 = c.get(c.S1, cache, n)
    S2 = c.get(c.S2, cache, n)

    gamma1_gq_hat = 2 * (
        16
        * CF
        * TR
        * (2 + n)
        * ((2 + 5 * n) / (9 * n * (1 + n) ** 2) - S1 / (3 * n * (1 + n)))
    )
    a_gq_l0 = (
        CF
        * TR
        * (
            (n + 2)
            * (
                8 * (22 + 41 * n + 28 * n**2) / (27 * n * (n + 1) ** 3)
                - 8 * (2 + 5 * n) / (9 * n * (n + 1) ** 2) * S1
                + ((S1**2 + S2) * 4) / (3 * n * (n + 1))
            )
        )
    )
    # here there is minus sing missing in the paper,
    # passing from eq 172 to 174
    a_gq_l1 = gamma1_gq_hat / 2
    a_gq_l2 = beta_0hat * gamma0_gq(n)

    return a_gq_l2 * L**2 + a_gq_l1 * (-L) + a_gq_l0


@nb.njit(cache=True)
def A_gg(n, cache, L):
    r"""Compute |NNLO| gluon-gluon |OME| :math:`A_{gg,H}^{S,(2)}`.

    Implements :eqref:`187` of :cite:`Bierenbaum:2022biv`.

    Parameters
    ----------
    n : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    complex
        |NNLO| gluon-gluon |OME| :math:`A_{gg,H}^{S,(2)}`

    """
    S1 = c.get(c.S1, cache, n)
    ggg1_canf = (
        -5 * S1 / 9
        + (-3 + 13 * n + 16 * n**2 + 6 * n**3 + 3 * n**4)
        / (9 * n**2 * (1 + n) ** 2)
    ) * 4
    ggg1_cfnf = (4 + 2 * n - 8 * n**2 + n**3 + 5 * n**4 + 3 * n**5 + n**6) / (
        n**3 * (1 + n) ** 3
    )
    gamma1_gg_hat = 8 * TR * (CA * ggg1_canf + CF * ggg1_cfnf)

    a_gg_f = (
        -15 * n**8
        - 60 * n**7
        - 82 * n**6
        - 44 * n**5
        - 15 * n**4
        - 4 * n**2
        - 12 * n
        - 8
    ) / (n**4 * (1 + n) ** 4)
    a_gg_a = (
        2
        * (
            15 * n**6
            + 45 * n**5
            + 374 * n**4
            + 601 * n**3
            + 161 * n**2
            - 24 * n
            + 36
        )
    ) / (27 * n**3 * (1 + n) ** 3) - (4 * S1 * (47 + 56 * n) / (27 * (1 + n)))
    a_gg_l0 = TR * (CF * a_gg_f + CA * a_gg_a)

    # here there is minus sing missing in the paper,
    # passing from eq 177 to 185
    a_gg_l1 = gamma1_gg_hat / 2

    # remove the nf dependence from
    # (2 * gamma0_gg(n, S1, nf) + 2 * beta_0(nf))
    pgg0_beta0 = 8 * CA * (-(2 / (n + n**2)) + S1)
    a_gg_l2 = (1 / 8) * (
        2 * beta_0hat * pgg0_beta0
        + 2 * gamma0_gq(n) * 2 * gamma0_qg(n, nf=1)
        + 8 * beta_0hat**2
    )
    return a_gg_l2 * L**2 + a_gg_l1 * (-L) + a_gg_l0


@nb.njit(cache=True)
def A_singlet(n, cache, L, nf):
    r"""Compute the |NNLO| singlet |OME|.

    .. math::
        A^{S,(2)} = \left(\begin{array}{cc}
        A_{gg, H}^{S,(2)} & A_{gq, H}^{S,(2)} & 0 \\
        0 & A_{qq,H}^{NS,(2)} & 0\\
        A_{hg}^{S,(2)} & A_{hq}^{PS,(2)} & 0\\
        \end{array}\right)

    Parameters
    ----------
    n : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        |NNLO| singlet |OME| :math:`A^{S,(2)}(N)`

    """
    return np.array(
        [
            [A_gg(n, cache, L), A_gq(n, cache, L), 0.0],
            [0.0, A_qq_ns(n, cache, L), 0.0],
            [A_hg(n, cache, L), A_hq_ps(n, cache, L, nf), 0.0],
        ],
        np.complex_,
    )


@nb.njit(cache=True)
def A_ns(n, cache, L):
    r"""Compute the |NNLO| non-singlet |OME|.

    .. math::
        A^{NS,(2)} = \left(\begin{array}{cc}
        A_{qq,H}^{NS,(2)} & 0 \\
        0 & 0 \\
        \end{array}\right)

    Parameters
    ----------
    n : complex
        Mellin moment
    cache: numpy.ndarray
        Harmonic sum cache
    L : float
        :math:`\ln(\mu_F^2 / m_h^2)`

    Returns
    -------
    numpy.ndarray
        |NNLO| non-singlet |OME| :math:`A^{NS,(2)}`

    """
    return np.array([[A_qq_ns(n, cache, L), 0.0], [0 + 0j, 0 + 0j]], np.complex_)



"""Collection of QED valence EKOs."""
import numba as nb

from .singlet_qed import eko_iterate


@nb.njit(cache=True)
def dispatcher(
    order,
    method,
    gamma_valence,
    as_list,
    a_half,
    nf,
    ev_op_iterations,
    ev_op_max_order,
):
    """
    Determine used kernel and call it.

    Parameters
    ----------
    order : tuple(int,int)
        perturbative order
    method : str
        method
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    aem_list : numpy.ndarray
        electromagnetic coupling values
    nf : int
        number of active flavors
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : tuple(int,int)
        perturbative expansion order of U

    Returns
    -------
        e_v : numpy.ndarray
            singlet EKO
    """
    if method in ["iterate-exact", "iterate-expanded"]:
        return eko_iterate(
            gamma_valence, as_list, a_half, nf, order, ev_op_iterations, 2
        )
    raise NotImplementedError('Only "iterate-exact" is implemented with QED')



r"""The write-up of the matching conditions is given in :doc:`Matching Conditions </theory/Matching>`."""

import numpy as np

from .. import basis_rotation as br
from .. import constants


def pids_from_intrinsic_evol(label, nf, normalize):
    r"""Obtain the list of pids with their corresponding weight, that are contributing to ``evol``.

    The normalization of the weights is only needed for the output rotation:

    - if we want to build e.g. the singlet in the initial state we simply have to sum
      to obtain :math:`\Sigma = u + \bar u + d + \bar d + \ldots`
    - if we want to rotate back in the output we have to *normalize* the weights:
      e.g. in nf=3 :math:`u = \frac 1 6 \Sigma + \frac 1 6 V + \ldots`

    The normalization can only happen here since we're actively cutting out some
    flavor (according to ``nf``).

    Parameters
    ----------
    label : str
        evolution label
    nf : int
        maximum number of light flavors
    normalize : bool
        normalize output

    Returns
    -------
    list(float)
        list of weights
    """
    try:
        evol_idx = br.evol_basis.index(label)
        is_evol = True
    except ValueError:
        is_evol = False
    if is_evol:
        weights = br.rotate_flavor_to_evolution[evol_idx].copy()
        for j, pid in enumerate(br.flavor_basis_pids):
            if nf < abs(pid) <= 6:
                weights[j] = 0
    else:
        weights = rotate_pm_to_flavor(label)
    # normalize?
    if normalize:
        norm = weights @ weights
        weights = weights / norm
    return weights


def get_range(evol_labels, qed=False):
    """Determine the number of light and heavy flavors participating in the input and output.

    Here, we assume that the T distributions (e.g. T15) appears *always*
    before the corresponding V distribution (e.g. V15).

    Parameters
    ----------
    qed : bool
        activate qed

    Returns
    -------
    nf_in : int
        number of light flavors in the input
    nf_out : int
        number of light flavors in the output
    """
    nf_in = 3
    nf_out = 3

    def update(label, qed=False):
        nf = 3
        if label[0] == "T":
            if not qed:
                nf = round(np.sqrt(int(label[1:]) + 1))
            else:
                if label[1:] == "d3":
                    nf = 3
                elif label[1:] == "u3":
                    nf = 4
                elif label[1:] == "d8":
                    nf = 5
                elif label[1:] == "u8":
                    nf = 6
                else:
                    raise ValueError(f"{label[1:]} is not possible")
        return nf

    for op in evol_labels:
        nf_in = max(update(op.input, qed), nf_in)
        nf_out = max(update(op.target, qed), nf_out)

    return nf_in, nf_out


def rotate_pm_to_flavor(label):
    """Rotate from +- basis to flavor basis.

    Parameters
    ----------
    label : str
        label

    Returns
    -------
    list(float)
        list of weights
    """
    # g and ph are unaltered
    if label in ["g", "ph"]:
        return br.rotate_flavor_to_evolution[br.evol_basis.index(label)].copy()
    # no it has to be a quark with + or - appended
    if label[0] not in br.quark_names or label[1] not in ["+", "-"]:
        raise ValueError(f"Invalid pm label: {label}")
    l = np.zeros(len(br.flavor_basis_pids))
    idx = br.flavor_basis_names.index(label[0])
    pid = br.flavor_basis_pids[idx]
    l[idx] = 1
    # + is +, - is -
    if label[1] == "+":
        l[br.flavor_basis_pids.index(-pid)] = 1
    else:
        l[br.flavor_basis_pids.index(-pid)] = -1
    return l


def rotate_matching(nf, qed=False, inverse=False):
    """Rotation between matching basis (with e.g. S,g,...V8 and c+,c-) and new true evolution basis (with S,g,...V8,T15,V15).

    Parameters
    ----------
    nf : int
        number of active flavors in the higher patch: to activate T15, nf=4
    qed : bool
        use QED?
    inverse : bool
        use inverse conditions?

    Returns
    -------
    dict
        mapping in dot notation between the bases
    """
    # the gluon and the photon do not care about new quarks
    l = {"g.g": 1.0, "ph.ph": 1.0}
    # already active distributions
    q = br.quark_names[nf - 1]
    if not qed:
        for k in range(2, nf):  # nf is the upper, so excluded
            n = k**2 - 1
            l[f"V{n}.V{n}"] = 1.0
            l[f"T{n}.T{n}"] = 1.0
        # the new contributions
        n = nf**2 - 1  # nf is pointing upwards
        for tot, oth, qpm in (("S", f"T{n}", f"{q}+"), ("V", f"V{n}", f"{q}-")):
            if inverse:
                l[f"{tot}.{tot}"] = (nf - 1.0) / nf
                l[f"{tot}.{oth}"] = 1.0 / nf
                l[f"{qpm}.{tot}"] = 1.0 / nf
                l[f"{qpm}.{oth}"] = -1.0 / nf
            else:
                l[f"{tot}.{tot}"] = 1.0
                l[f"{tot}.{qpm}"] = 1.0
                l[f"{oth}.{tot}"] = 1.0
                l[f"{oth}.{qpm}"] = -(nf - 1.0)
    else:
        names = {3: "d3", 4: "u3", 5: "d8", 6: "u8"}
        for k in range(3, nf):
            l[f"V{names[k]}.V{names[k]}"] = 1.0
            l[f"T{names[k]}.T{names[k]}"] = 1.0
        for tot, totdelta, oth, qpm in (
            ("S", "Sdelta", f"T{names[nf]}", f"{q}+"),
            ("V", "Vdelta", f"V{names[nf]}", f"{q}-"),
        ):
            a, b, c, d, e, f = qed_rotation_parameters(nf)
            if inverse:
                den = -b * d + a * e - c * e + b * f
                l[f"{tot}.{tot}"] = -(c * e - b * f) / den
                l[f"{tot}.{totdelta}"] = e / den
                l[f"{tot}.{oth}"] = -b / den
                l[f"{totdelta}.{tot}"] = (c * d - a * f) / den
                l[f"{totdelta}.{totdelta}"] = (f - d) / den
                l[f"{totdelta}.{oth}"] = (a - c) / den
                l[f"{qpm}.{tot}"] = (-b * d + a * e) / den
                l[f"{qpm}.{totdelta}"] = -e / den
                l[f"{qpm}.{oth}"] = b / den
            else:
                l[f"{tot}.{tot}"] = 1.0
                l[f"{tot}.{qpm}"] = 1.0
                l[f"{totdelta}.{tot}"] = a
                l[f"{totdelta}.{totdelta}"] = b
                l[f"{totdelta}.{qpm}"] = c
                l[f"{oth}.{tot}"] = d
                l[f"{oth}.{totdelta}"] = e
                l[f"{oth}.{qpm}"] = f
    # also higher quarks do not care
    for k in range(nf + 1, 6 + 1):
        q = br.quark_names[k - 1]
        for sgn in "+-":
            l[f"{q}{sgn}.{q}{sgn}"] = 1.0
    return l


def rotate_matching_inverse(nf, qed=False):
    """Inverse rotation between matching basis (with e.g. S,g,...V8 and c+,c-) and new true evolution basis (with S,g,...V8,T15,V15).

    Parameters
    ----------
    nf : int
        number of active flavors in the higher patch: to activate T15, nf=4
    qed : bool
        use QED?

    Returns
    -------
    dict
        mapping in dot notation between the bases
    """
    return rotate_matching(nf, qed, True)


def qed_rotation_parameters(nf):
    r"""Parameters of the QED basis rotation.

    From :math:`(\Sigma, \Sigma_{\Delta}, h_+)` into :math:`(\Sigma, \Sigma_{\Delta}, T_i^j)`,
    or equivalentely for :math:`V, V_{\Delta}, V_i^j, h_-`.

    Parameters
    ----------
    nf : int
        number of active flavors in the higher patch: e.g. to activate :math:`T_3^u` or :math:`V_3^u` choose ``nf=4``

    Returns
    -------
    a,b,c,d,e,f : float
        Parameters of the rotation: :math:`\Sigma_{\Delta} = a*\Sigma + b*\Sigma_{\Delta} + c*h_+, T_i = d*\Sigma + e*\Sigma_{\Delta} + f*h_+`
    """
    nu_l = constants.uplike_flavors(nf - 1)
    nd_l = (nf - 1) - nu_l
    nu_h = constants.uplike_flavors(nf)
    nd_h = nf - nu_h
    a = (nd_h / nu_h * nu_l - nd_l) / (nf - 1)
    b = nf / nu_h * nu_l / (nf - 1)
    if nf in [4, 6]:  # heavy flavor is up-like
        c = nd_h / nu_h
        d = nu_l / (nf - 1)
        e = nu_l / (nf - 1)
    elif nf in [3, 5]:  # heavy flavor is down-like
        c = -1
        d = nd_l / (nf - 1)
        e = -nu_l / (nf - 1)
    if nf in [3, 4]:  # s and c unlock T3d, T3u that have -h+
        f = -1
    elif nf in [5, 6]:  # b and t unlock T8d, T8u that have -2h+
        f = -2
    return a, b, c, d, e, f


def pids_from_intrinsic_unified_evol(label, nf, normalize):
    r"""Obtain the list of pids with their corresponding weight, that are contributing to intrinsic unified evolution.

    Parameters
    ----------
    evol : str
        evolution label
    nf : int
        maximum number of light flavors
    normalize : bool
        normalize output

    Returns
    -------
    list(float)
        list of weights
    """
    if label in ["ph", "g", "S", "V"]:
        return pids_from_intrinsic_evol(label, nf, normalize)
    if label[0] in br.quark_names[3:]:
        weights = rotate_pm_to_flavor(label)
    else:
        weights = np.array([0.0] * len(br.flavor_basis_pids))
        mapping = {
            "d3": {1: 1.0, 3: -1.0},  # T3d = d+ - s+
            "d8": {1: 1.0, 3: 1.0, 5: -2.0},  # T8d = d+ + s+ - 2b+
            "u3": {2: 1.0, 4: -1.0},  # T3u = u+ - c+
            "u8": {2: 1.0, 4: 1.0, 6: -2.0},  # T8u = u+ + c+ - 2t+
            "delta": {
                3: {2: 2.0, 1: -1.0, 3: -1.0},  # Sdelta = 2u+ - d+ -s+
                4: {2: 1.0, 4: 1.0, 1: -1.0, 3: -1.0},  # Sdelta = u+ + c+ - d+ - s+
                5: {
                    2: 3.0 / 2.0,
                    4: 3.0 / 2.0,
                    1: -1.0,
                    3: -1.0,
                    5: -1.0,
                },  # Sdelta = 3/2u+ + 3/2c+ - d+ - s+ - b+
                6: {
                    2: 1.0,
                    4: 1.0,
                    6: 1.0,
                    1: -1.0,
                    3: -1.0,
                    5: -1.0,
                },  # Sdelta = u+ + c+ + t+ - d+ -s+ - b+
            },
        }
        cur_map = mapping[label[1:]]
        if label[1:] == "delta":
            cur_map = cur_map[nf]
        for q, w in cur_map.items():
            weights[br.flavor_basis_pids.index(q)] = w
            weights[br.flavor_basis_pids.index(-q)] = (
                -1 if label[0] == "V" else 1.0
            ) * w

    # normalize?
    if normalize:
        norm = weights @ weights
        weights = weights / norm
    return weights



"""Weight 2 harmonic sums."""

import numba as nb

from eko.constants import zeta2

from .polygamma import cern_polygamma


@nb.njit(cache=True)
def S2(N):
    r"""Compute the harmonic sum :math:`S_2(N)`.

    .. math::
      S_2(N) = \sum\limits_{j=1}^N \frac 1 {j^2} = -\psi_1(N+1)+\zeta(2)

    with :math:`\psi_1(N)` the trigamma function and :math:`\zeta` the
    Riemann zeta function.

    Parameters
    ----------
    N : complex
        Mellin moment

    Returns
    -------
    S_2 : complex
        Harmonic sum :math:`S_2(N)`

    See Also
    --------
        ekore.harmonics.polygamma.cern_polygamma : :math:`\psi_k(N)`

    """
    return -cern_polygamma(N + 1.0, 1) + zeta2


@nb.njit(cache=True)
def Sm2(N, hS2, hS2mh, hS2h, is_singlet=None):
    r"""Analytic continuation of harmonic sum :math:`S_{-2}(N)`.

    .. math::
      S_{-2}(N) = \sum\limits_{j=1}^N \frac {(-1)^j}{j^2}

    Parameters
    ----------
    N : complex
        Mellin moment
    hS2:  complex
        Harmonic sum :math:`S_{2}(N)`
    hS2mh: complex
        Harmonic sum :math:`S_{2}((N-1)/2)`
    hS2h: complex
        Harmonic sum :math:`S_{2}(N/2)`
    is_singlet: bool, None
        symmetry factor: True for singlet like quantities
        (:math:`\eta=(-1)^N = 1`), False for non-singlet like quantities
        (:math:`\eta=(-1)^N=-1`)

    Returns
    -------
    Sm2 : complex
        Harmonic sum :math:`S_{-2}(N)`

    See Also
    --------
        eko.anomalous_dimension.w2.S2 : :math:`S_2(N)`

    """
    if is_singlet is None:
        return 1 / 2 * ((1 - (-1) ** N) / 2 * hS2mh + ((-1) ** N + 1) / 2 * hS2h) - hS2
    if is_singlet:
        return 1 / 2 * hS2h - hS2
    return 1 / 2 * hS2mh - hS2



"""Manage assets used during computation."""
import base64
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Dict, Generic, Optional, Type, TypeVar

import yaml

from .access import AccessConfigs
from .items import Header, Operator

NBYTES = 8
ENDIANNESS = "little"

HEADER_EXT = ".yaml"
ARRAY_EXT = [".npy", ".npz"]
COMPRESSED_EXT = ".lz4"
OPERATOR_EXT = [ext + COMPRESSED_EXT for ext in ARRAY_EXT]


class LookupError(ValueError):
    """Failure in content retrieval from inventory."""


def encode(header: Header):
    """Extract an hash from a header."""
    return base64.urlsafe_b64encode(
        abs(hash(header)).to_bytes(NBYTES, byteorder=ENDIANNESS)
    ).decode(encoding="utf-8")


def header_name(header: Header):
    """Determine header file name."""
    stem = encode(header)
    return stem + HEADER_EXT


def operator_name(header: Header, err: bool):
    """Determine operator file name, from the associated header."""
    stem = encode(header)
    return stem + OPERATOR_EXT[1 if err else 0]


H = TypeVar("H", bound=Header)


@dataclass(frozen=True)
class Inventory(Generic[H]):
    """Assets manager.

    In particular, manage autosave, autoload, and memory caching.

    """

    path: Path
    access: AccessConfigs
    header_type: Type[Header]
    cache: Dict[H, Optional[Operator]] = field(default_factory=dict)
    contentless: bool = False
    name: Optional[str] = None
    """Only for logging purpose."""

    def __str__(self) -> str:
        return f"Inventory '{self.name}'"

    def lookup(self, stem: str, header: bool = False) -> Path:
        """Look up for content path in inventory."""
        EXT = OPERATOR_EXT if not header else [HEADER_EXT]
        found = [
            path
            for path in self.path.iterdir()
            if path.name.startswith(stem)
            if "".join(path.suffixes) in EXT
        ]

        if len(found) == 0:
            raise LookupError(f"Item '{stem}' not available in {self}.")
        elif len(found) > 1:
            raise LookupError(
                f"Too many items associated to '{stem}' in {self}:\n{found}"
            )

        return found[0]

    def __getitem__(self, header: H) -> Optional[Operator]:
        r"""Retrieve operator for given header.

        If the operator is not already in memory, it will be automatically
        loaded.

        """
        self.access.assert_open()

        try:
            op = self.cache[header]
            if op is not None or self.contentless:
                return op
        except KeyError:
            pass

        stem = encode(header)

        # in case of contentless, check header availability instead
        if self.contentless:
            self.lookup(stem, header=True)
            self.cache[header] = None
            return None

        # for contentful inventories, check operator availability
        oppath = self.lookup(stem)

        with open(oppath, "rb") as fd:
            op = Operator.load(fd)

        self.cache[header] = op
        return op

    def __setitem__(self, header: H, operator: Optional[Operator]):
        """Set operator for given header.

        Header and operator are automatically dumped on disk.

        """
        self.access.assert_writeable()

        # always save the header on disk
        headpath = self.path / header_name(header)
        headpath.write_text(yaml.dump(asdict(header)), encoding="utf-8")

        # in case of contentless, set empty cache and exit
        if self.contentless:
            self.cache[header] = None
            return

        # otherwise save also the operator, and add to the cache
        assert operator is not None

        with_err = operator.error is not None
        oppath = self.path / operator_name(header, err=with_err)

        with open(oppath, "wb") as fd:
            operator.save(fd)

        self.cache[header] = operator

    def __delitem__(self, header: H):
        """Drop operator from memory.

        Irrelevant for contentless inventories.

        Note
        ----
        This method only drops the operator from memory, and it's not expected
        to do anything else.

        Autosave is done on set, and explicit saves are performed by the
        computation functions.

        If a further explicit save is required, repeat explicit assignment::

            inventory[header] = inventory[header]

        This is only useful if the operator has been mutated in place, that in
        general should be avoided, since the operator should only be the result
        of a full computation or a library manipulation.

        """
        self.cache[header] = None

    def __iter__(self):
        """Iterate over loaded content.

        This iteration is only over cache, so it might not be faithful with
        respect to the real content on disk.
        To iterate the full content of the disk, just call right before
        :meth:`sync`.

        """
        yield from self.cache

    def __len__(self):
        """Return the number of elements in the cache."""
        return len(self.cache)

    def sync(self):
        """Sync the headers in the cache with the content on disk.

        In particular, headers on disk that are missing in the :attr:`cache`
        are added to it, without loading actual operators in memory.

        Despite the name, the operation is non-destructive, so, even if cache
        has been abused, nothing will be deleted nor unloaded.

        """
        for path in self.path.iterdir():
            if path.suffix != HEADER_EXT:
                continue

            header = self.header_type(
                **yaml.safe_load(path.read_text(encoding="utf-8"))
            )
            self.cache[header] = None

    def __invert__(self):
        """Alias for :meth:`sync`."""
        self.sync()

    def empty(self):
        """Empty the in-memory cache."""
        for header in self.cache:
            del self[header]



"""Collection of QED singlet EKOs."""
import numba as nb
import numpy as np

from ekore import anomalous_dimensions as ad

from .. import beta


@nb.njit(cache=True)
def eko_iterate(gamma_singlet, as_list, a_half, nf, order, ev_op_iterations, dim):
    """Singlet QEDxQCD iterated (exact) EKO.

    Parameters
    ----------
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target strong coupling value
    a0 : float
        initial strong coupling value
    aem_list : float
        electromagnetic coupling values
    nf : int
        number of active flavors
    order : tuple(int,int)
        QCDxQED perturbative orders
    ev_op_iterations : int
        number of evolution steps

    Returns
    -------
    e_s^{order} : numpy.ndarray
        singlet QEDxQCD iterated (exact) EKO
    """
    e = np.identity(dim, np.complex_)
    betaQCD = np.zeros((order[0] + 1, order[1] + 1))
    for i in range(1, order[0] + 1):
        betaQCD[i, 0] = beta.beta_qcd((i + 1, 0), nf)
    betaQCD[1, 1] = beta.beta_qcd((2, 1), nf)
    for step in range(1, ev_op_iterations + 1):
        ah = as_list[step]
        al = as_list[step - 1]
        as_half = a_half[step - 1, 0]
        aem = a_half[step - 1, 1]
        delta_a = ah - al
        gamma = np.zeros((dim, dim), np.complex_)
        betatot = 0
        for i in range(0, order[0] + 1):
            for j in range(0, order[1] + 1):
                betatot += betaQCD[i, j] * as_half ** (i + 1) * aem**j
                gamma += gamma_singlet[i, j] * as_half**i * aem**j
        ln = gamma / betatot * delta_a
        ek = np.ascontiguousarray(ad.exp_matrix(ln)[0])
        e = ek @ e
    return e


@nb.njit(cache=True)
def dispatcher(
    order,
    method,
    gamma_singlet,
    as_list,
    a_half,
    nf,
    ev_op_iterations,
    ev_op_max_order,
):
    """Determine used kernel and call it.

    Parameters
    ----------
    order : tuple(int,int)
        perturbative order
    method : str
        method
    gamma_singlet : numpy.ndarray
        singlet anomalous dimensions matrices
    a1 : float
        target coupling value
    a0 : float
        initial coupling value
    aem_list : numpy.ndarray
        electromagnetic coupling values
    nf : int
        number of active flavors
    ev_op_iterations : int
        number of evolution steps
    ev_op_max_order : tuple(int,int)
        perturbative expansion order of U

    Returns
    -------
    e_s : numpy.ndarray
        singlet EKO
    """
    if method in ["iterate-exact", "iterate-expanded"]:
        return eko_iterate(
            gamma_singlet, as_list, a_half, nf, order, ev_op_iterations, 4
        )
    raise NotImplementedError('Only "iterate-exact" is implemented with QED')



r"""The polarized, space-like Altarelli-Parisi splitting kernels.

Normalization is given by

.. math::
    \mathbf{P}(x) = \sum\limits_{j=0} a_s^{j+1} \mathbf P^{(j)}(x)

with :math:`a_s = \frac{\alpha_S(\mu^2)}{4\pi}`.
"""

import numba as nb
import numpy as np

from ....harmonics import cache as c
from . import as1, as2, as3


@nb.njit(cache=True)
def gamma_ns(order, mode, n, nf):
    r"""Compute the tower of the non-singlet anomalous dimensions.

    Parameters
    ----------
    order : tuple(int,int)
        perturbative orders
    mode : 10201 | 10101 | 10200
        sector identifier
    n : complex
        Mellin variable
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        non-singlet anomalous dimensions

    """
    cache = c.reset()
    gamma_ns = np.zeros(order[0], np.complex_)
    gamma_ns[0] = as1.gamma_ns(n, cache)
    # NLO and beyond
    if order[0] >= 2:
        if mode == 10101:
            gamma_ns_1 = as2.gamma_nsp(n, nf, cache)
        # To fill the full valence vector in NNLO we need to add gamma_ns^1 explicitly here
        elif mode in [10201, 10200]:
            gamma_ns_1 = as2.gamma_nsm(n, nf, cache)
        else:
            raise NotImplementedError("Non-singlet sector is not implemented")
        gamma_ns[1] = gamma_ns_1
    if order[0] >= 3:
        if mode == 10101:
            gamma_ns_2 = as3.gamma_nsp(n, nf, cache)
        elif mode == 10201:
            gamma_ns_2 = as3.gamma_nsm(n, nf, cache)
        elif mode == 10200:
            gamma_ns_2 = as3.gamma_nsv(n, nf, cache)
        gamma_ns[2] = gamma_ns_2
    if order[0] >= 4:
        raise NotImplementedError("Polarized beyond NNLO is not yet implemented")
    return gamma_ns


@nb.njit(cache=True)
def gamma_singlet(order, n, nf):
    r"""Compute the tower of the singlet anomalous dimensions matrices.

    Parameters
    ----------
    order : tuple(int,int)
        perturbative orders
    n : complex
        Mellin variable
    nf : int
        Number of active flavors

    Returns
    -------
    numpy.ndarray
        singlet anomalous dimensions matrices

    """
    cache = c.reset()
    gamma_s = np.zeros((order[0], 2, 2), np.complex_)
    gamma_s[0] = as1.gamma_singlet(n, cache, nf)
    if order[0] >= 2:
        gamma_s[1] = as2.gamma_singlet(n, nf, cache)
    if order[0] >= 3:
        gamma_s[2] = as3.gamma_singlet(n, nf, cache)
    if order[0] >= 4:
        raise NotImplementedError("Polarized beyond NNLO is not yet implemented")
    return gamma_s



"""Contains the :class:`PhysicalOperator` class."""
from .. import basis_rotation as br
from .. import member


class PhysicalOperator(member.OperatorBase):
    """
    Join several fixed flavor scheme operators together.

    - provides the connection between the 7-dimensional anomalous dimension
      basis and the 15-dimensional evolution basis
    - provides the connection between the 15-dimensional evolution basis
      and the 169-dimensional flavor basis


    Parameters
    ----------
        op_members : dict
            list of all members
        q2_final : float
            final scale
    """

    @classmethod
    def ad_to_evol_map(cls, op_members, nf, q2_final, intrinsic_range, qed=False):
        """
        Obtain map between the 3-dimensional anomalous dimension basis and the 4-dimensional evolution basis.

        .. todo:: in VFNS sometimes IC is irrelevant if nf>=4

        Parameters
        ----------
            op_members : dict
                operator members in anomalous dimension basis
            nf : int
                number of active light flavors
            intrinsic_range : sequence
                intrinsic heavy flavors
            qed : bool
                activate qed

        Returns
        -------
            m : dict
                map
        """
        # constant elements
        m = {
            "S.S": op_members[(100, 100)],
            "S.g": op_members[(100, 21)],
            "g.g": op_members[(21, 21)],
            "g.S": op_members[(21, 100)],
        }
        if not qed:
            m.update({"V.V": op_members[(br.non_singlet_pids_map["nsV"], 0)]})
            # add elements which are already active
            for f in range(2, nf + 1):
                n = f**2 - 1
                m[f"V{n}.V{n}"] = op_members[(br.non_singlet_pids_map["ns-"], 0)]
                m[f"T{n}.T{n}"] = op_members[(br.non_singlet_pids_map["ns+"], 0)]
        else:
            m.update(
                {
                    "g.ph": op_members[(21, 22)],
                    "g.Sdelta": op_members[(21, 101)],
                    "ph.g": op_members[(22, 21)],
                    "ph.ph": op_members[(22, 22)],
                    "ph.S": op_members[(22, 100)],
                    "ph.Sdelta": op_members[(22, 101)],
                    "S.ph": op_members[(100, 22)],
                    "S.Sdelta": op_members[(100, 101)],
                    "Sdelta.g": op_members[(101, 21)],
                    "Sdelta.ph": op_members[(101, 22)],
                    "Sdelta.S": op_members[(101, 100)],
                    "Sdelta.Sdelta": op_members[(101, 101)],
                    "V.V": op_members[(10200, 10200)],
                    "V.Vdelta": op_members[(10200, 10204)],
                    "Vdelta.V": op_members[(10204, 10200)],
                    "Vdelta.Vdelta": op_members[(10204, 10204)],
                }
            )
            # add elements which are already active
            if nf >= 3:
                m["Td3.Td3"] = op_members[(br.non_singlet_pids_map["ns+d"], 0)]
                m["Vd3.Vd3"] = op_members[(br.non_singlet_pids_map["ns-d"], 0)]
            if nf >= 4:
                m["Tu3.Tu3"] = op_members[(br.non_singlet_pids_map["ns+u"], 0)]
                m["Vu3.Vu3"] = op_members[(br.non_singlet_pids_map["ns-u"], 0)]
            if nf >= 5:
                m["Td8.Td8"] = op_members[(br.non_singlet_pids_map["ns+d"], 0)]
                m["Vd8.Vd8"] = op_members[(br.non_singlet_pids_map["ns-d"], 0)]
            if nf >= 6:
                m["Tu8.Tu8"] = op_members[(br.non_singlet_pids_map["ns+u"], 0)]
                m["Vu8.Vu8"] = op_members[(br.non_singlet_pids_map["ns-u"], 0)]
        # deal with intrinsic heavy quark PDFs
        if intrinsic_range is not None:
            hqfl = "cbt"
            op_id = member.OpMember.id_like(op_members[(21, 21)])
            for intr_fl in intrinsic_range:
                if intr_fl <= nf:  # light quarks are not intrinsic
                    continue
                hq = hqfl[intr_fl - 4]  # find name
                # intrinsic means no evolution, i.e. they are evolving with the identity
                m[f"{hq}+.{hq}+"] = op_id.copy()
                m[f"{hq}-.{hq}-"] = op_id.copy()
        # map key to MemberName
        return cls.promote_names(m, q2_final)



"""Abstract layer for running the benchmarks."""
import functools
import logging
import os
import sys

import pandas as pd
from banana import cfg as banana_cfg
from banana.benchmark.runner import BenchmarkRunner
from banana.data import dfdict

import eko
from eko import EKO
from eko import basis_rotation as br
from eko.io import manipulate
from ekobox import apply

from .. import pdfname
from ..data import db, operators


class Runner(BenchmarkRunner):
    """EKO specialization of the banana runner."""

    db_base_cls = db.Base
    rotate_to_evolution_basis = False
    sandbox = False
    plot_operator = False
    log_to_stdout = True

    def __init__(self):
        self.banana_cfg = banana_cfg.cfg

    @staticmethod
    def load_ocards(session, ocard_updates):
        """
        Load operator cards.

        Parameters
        ----------
        session : sqlalchemy.session.Session
            DB ORM session
        updates : dict
            modifiers

        Returns
        -------
        cards : list(dict)
            list of records
        """
        return operators.load(session, ocard_updates)

    @staticmethod
    def skip_pdfs(_theory):
        """Specify PDFs to skip."""
        return []

    def run_me(self, theory, ocard, _pdf):
        """
        Run eko.

        Parameters
        ----------
        theory : dict
            theory card
        ocard : dict
            operator card
        pdf : lhapdf_type
            pdf

        Returns
        -------
            out :  dict
                DGLAP result
        """
        # activate logging
        if self.log_to_stdout:
            logStdout = logging.StreamHandler(sys.stdout)
            logStdout.setLevel(logging.INFO)
            logStdout.setFormatter(logging.Formatter("%(message)s"))
            logging.getLogger("eko").handlers = []
            logging.getLogger("eko").addHandler(logStdout)
            logging.getLogger("eko").setLevel(logging.INFO)

        ops_id = f"o{ocard['hash'][:6]}_t{theory['hash'][:6]}"
        root = banana_cfg.cfg["paths"]["database"].parents[0]
        path = root / f"{ops_id}.tar"

        # if sandbox check for cache, dump eko to yaml
        # and plot the operators
        if self.sandbox:
            rerun = True

            if os.path.exists(path):
                rerun = False
                ask = input("Use cached output? [Y/n]")
                if ask.lower() in ["n", "no"]:
                    rerun = True

            if rerun:
                path.unlink(missing_ok=True)
                eko.solve(theory, ocard, path)
                print(f"Operator written to {path}")
            else:
                # load
                print(f"Using cached eko data: {os.path.relpath(path,os.getcwd())}")

            if self.plot_operator:
                from ekomark.plots import (  # pylint:disable=import-error,import-outside-toplevel
                    save_operators_to_pdf,
                )

                output_path = f"{root}/{self.external}_bench"
                if not os.path.exists(output_path):
                    os.makedirs(output_path)
                # rotating to evolution basis if requested
                with EKO.edit(path) as out_copy:
                    change_lab = False
                    if self.rotate_to_evolution_basis:
                        qed = theory["QED"] > 0
                        if not qed:
                            manipulate.to_evol(out_copy, source=True, target=True)
                        else:
                            manipulate.to_uni_evol(out_copy, source=True, target=True)
                        change_lab = True

                    save_operators_to_pdf(
                        output_path,
                        theory,
                        ocard,
                        out_copy,
                        self.skip_pdfs(theory),
                        change_lab,
                    )
        else:
            # else we always rerun
            path.unlink(missing_ok=True)
            eko.solve(theory, ocard, path)
        return path

    def run_external(self, theory, ocard, pdf):
        """
        Run external library.

        Parameters
        ----------
            theory : dict
                theory card
            ocard : dict
                operator card
            pdf : lhapdf_type
                pdf

        Returns
        -------
            out :  dict
                DGLAP result
        """
        # pylint:disable=import-error,import-outside-toplevel
        if self.external.lower() == "lha":
            from .external import LHA_utils

            # here pdf and skip_pdfs is not needed
            return LHA_utils.compute_LHA_data(
                theory,
                ocard,
                rotate_to_evolution_basis=self.rotate_to_evolution_basis,
            )
        if self.external.lower() == "lhapdf":
            from .external import lhapdf_utils

            # here theory card is not needed
            return lhapdf_utils.compute_LHAPDF_data(
                theory,
                ocard,
                pdf,
                self.skip_pdfs(theory),
                rotate_to_evolution_basis=self.rotate_to_evolution_basis,
            )
        if self.external.lower() == "pegasus":
            from .external import pegasus_utils

            return pegasus_utils.compute_pegasus_data(
                theory,
                ocard,
                self.skip_pdfs(theory),
                rotate_to_evolution_basis=self.rotate_to_evolution_basis,
            )

        if self.external.lower() == "apfel":
            from .external import apfel_utils

            return apfel_utils.compute_apfel_data(
                theory,
                ocard,
                pdf,
                self.skip_pdfs(theory),
                rotate_to_evolution_basis=self.rotate_to_evolution_basis,
            )

        raise NotImplementedError(
            f"Benchmark against {self.external} is not implemented!"
        )

    def log(self, theory, _, pdf, me, ext):
        """
        Return a proper log table.

        Parameters
        ----------
            theory : dict
                theory card
            pdf : lhapdf_type
                pdf
            me : eko.output.Output
                eko output object containing all informations

        """
        # return a proper log table
        log_tabs = {}
        xgrid = ext["target_xgrid"]
        q2s = list(ext["values"].keys())

        # LHA NNLO VFNS needs a special treatment
        # Valence contains only u and d
        rotate_to_evolution = None
        qed = theory["QED"] > 0
        if self.rotate_to_evolution_basis:
            if not qed:
                rotate_to_evolution = br.rotate_flavor_to_evolution.copy()
            else:
                rotate_to_evolution = br.rotate_flavor_to_unified_evolution.copy()
            if (
                self.external == "LHA"
                and theory["PTO"] == 2
                and theory["FNS"] == "ZM-VFNS"
            ):
                rotate_to_evolution[3, :] = [0, 0, 0, 0, 0, -1, -1, 0, 1, 1, 0, 0, 0, 0]

        with EKO.read(me) as eko:
            pdf_grid = apply.apply_pdf_flavor(
                eko,
                pdf,
                xgrid,
                flavor_rotation=rotate_to_evolution,
                qed=qed,
            )
        for q2 in q2s:
            log_tab = dfdict.DFdict()
            ref_pdfs = ext["values"][q2]
            res = pdf_grid[q2]
            my_pdfs = res["pdfs"]
            my_pdf_errs = res["errors"]

            for key in my_pdfs:
                if key in self.skip_pdfs(theory):
                    continue

                # build table
                tab = {}
                tab["x"] = xgrid
                tab["Q2"] = q2
                tab["eko"] = f = xgrid * my_pdfs[key]
                tab["eko_error"] = xgrid * my_pdf_errs[key]
                tab[self.external] = r = ref_pdfs[key]
                tab["percent_error"] = (f - r) / r * 100

                log_tab[pdfname(key)] = pd.DataFrame(tab)
            log_tabs[q2] = log_tab

        return functools.reduce(lambda dfd1, dfd2: dfd1.merge(dfd2), log_tabs.values())



"""Subcommand to manage runcards."""
import logging
import pathlib

from .. import cards
from . import library as lib
from .base import command

_logger = logging.getLogger(__name__)

DESTINATION = pathlib.Path.cwd().absolute() / "runcards"
"""Default destination for generated files"""

option_dest = lib.destination(DESTINATION)


@command.group("runcards")
def subcommand():
    """Manage EKO runcards."""


@subcommand.command("example")
@option_dest
def sub_example(destination: pathlib.Path):
    """Generate example runcards.

    Pay attention that they are in no way intended as defaults, but just
    examples to quickstart.

    """
    destination.mkdir(parents=True, exist_ok=True)
    theory = cards.example.theory()
    theory.order = (1, 0)
    cards.dump(theory.raw, path=destination / "theory.yaml")
    operator = cards.example.operator()
    operator.mu0 = 1.65
    operator.mu2grid = [1e5]
    cards.dump(operator.raw, path=destination / "operator.yaml")
    _logger.info(f"Runcards generated to '{destination}'")



r"""Manage running (and fixed) couplings.

Manage QCD coupling :math:`\alpha_s` and QED coupling :math:`\alpha`.
We provide an interface to access them simultaneously and provide several
strategies to solve the associated |RGE|.

See :doc:`pQCD ingredients </theory/pQCD>`.

"""
import logging
from typing import Iterable, List

import numba as nb
import numpy as np
import scipy

from . import constants, matchings
from .beta import b_qcd, b_qed, beta_qcd, beta_qed
from .io.types import EvolutionMethod, Order, SquaredScale
from .quantities.couplings import CouplingEvolutionMethod, CouplingsInfo
from .quantities.heavy_quarks import QuarkMassScheme

logger = logging.getLogger(__name__)


def couplings_mod_ev(mod_ev: EvolutionMethod) -> CouplingEvolutionMethod:
    """Map ModEv key to the available strong coupling evolution methods.

    Parameters
    ----------
    mod_ev : str
        evolution mode

    Returns
    -------
    str
        coupling mode
    """
    if mod_ev.value in [
        "EXA",
        "iterate-exact",
        "decompose-exact",
        "perturbative-exact",
    ]:
        return CouplingEvolutionMethod.EXACT
    if mod_ev.value in [
        "TRN",
        "truncated",
        "ordered-truncated",
        "EXP",
        "iterate-expanded",
        "decompose-expanded",
        "perturbative-expanded",
    ]:
        return CouplingEvolutionMethod.EXPANDED
    raise ValueError(f"Unknown evolution mode {mod_ev}")


@nb.njit(cache=True)
def exact_lo(ref, beta0, lmu):
    r"""Compute expanded solution at |LO|.

    Parameters
    ----------
    ref : float
        reference value of the coupling
    beta0 : float
        first coefficient of the beta function
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        coupling at target scale :math:`a(\mu_R^2)`
    """
    den = 1.0 + beta0 * ref * lmu
    return ref / den


@nb.njit(cache=True)
def expanded_nlo(ref, beta0, b1, lmu):
    r"""Compute expanded solution at |NLO|.

    Implement the default expression for |NLO| expanded solution, e.g. the one
    implemented in |APFEL|.

    Parameters
    ----------
    ref : float
        reference value of the coupling
    beta0 : float
        first coefficient of the beta function
    b1 : float
        second coefficient of the b function
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        coupling at target scale :math:`a(\mu_R^2)`
    """
    den = 1.0 + beta0 * ref * lmu
    a_LO = exact_lo(ref, beta0, lmu)
    as_NLO = a_LO * (1 - b1 * a_LO * np.log(den))
    return as_NLO


@nb.njit(cache=True)
def expanded_nnlo(ref, beta0, b1, b2, lmu):
    r"""Compute expanded solution at |NNLO|.

    Implement the |NNLO| expanded solution from |APFEL| (not the default expression)

    Parameters
    ----------
    ref : float
        reference value of the coupling
    beta0 : float
        first coefficient of the beta function
    b1 : float
        second coefficient of the b function
    b2 : float
        third coefficient of the b function
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        coupling at target scale :math:`a(\mu_R^2)`
    """
    a_LO = exact_lo(ref, beta0, lmu)
    a_NLO = expanded_nlo(ref, beta0, b1, lmu)
    res = a_LO * (
        1.0 + a_LO * (a_LO - ref) * (b2 - b1**2) + a_NLO * b1 * np.log(a_NLO / ref)
    )
    return res


@nb.njit(cache=True)
def expanded_n3lo(ref, beta0, b1, b2, b3, lmu):
    r"""Compute expanded solution at |N3LO|.

    Implement the |N3LO| expanded solution obtained via iterated solution of the RGE :cite:`Rottoli`

    Parameters
    ----------
    ref : float
        reference value of the coupling
    beta0 : float
        first coefficient of the beta function
    b1 : float
        second coefficient of the b function
    b2 : float
        third coefficient of the b function
    b3 : float
        fourth coefficient of the b function
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        coupling at target scale :math:`a(\mu_R^2)`
    """
    a_LO = exact_lo(ref, beta0, lmu)
    log_fact = np.log(a_LO)
    res = expanded_nnlo(ref, beta0, b1, b2, lmu)
    res += (
        a_LO**4
        / (2 * beta0**3)
        * (
            -2 * b1**3 * np.log(ref) ** 3
            + 5 * b1**3 * log_fact**2
            + 2 * b1**3 * log_fact**3
            + b1**3 * np.log(ref) ** 2 * (5 + 6 * log_fact)
            + 2 * beta0 * b1 * log_fact * (b2 + 2 * (b1**2 - beta0 * b2) * lmu * ref)
            - beta0**2
            * lmu
            * ref
            * (
                -2 * b1 * b2
                + 2 * beta0 * b3
                + (b1**3 - 2 * beta0 * b1 * b2 + beta0**2 * b3) * lmu * ref
            )
            - 2
            * b1
            * np.log(ref)
            * (
                5 * b1**2 * log_fact
                + 3 * b1**2 * log_fact**2
                + beta0 * (b2 + 2 * (b1**2 - beta0 * b2) * lmu * ref)
            )
        )
    )
    return res


@nb.njit(cache=True)
def expanded_qcd(ref, order, beta0, b_vec, lmu):
    r"""Compute QCD expanded solution at a given order.

    Parameters
    ----------
    ref : float
        reference value of the strong coupling
    order : int
        QCD order
    beta0 : float
        first coefficient of the beta function
    b_vec : list
        list of b function coefficients (including b0)
    lmu : float
        logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        strong coupling at target scale :math:`a_s(\mu_R^2)`
    """
    res_as = ref
    # LO
    if order == 1:
        res_as = exact_lo(ref, beta0, lmu)
    # NLO
    if order == 2:
        res_as = expanded_nlo(ref, beta0, b_vec[1], lmu)
    # NNLO
    if order == 3:
        res_as = expanded_nnlo(ref, beta0, b_vec[1], b_vec[2], lmu)
    # N3LO
    if order == 4:
        res_as = expanded_n3lo(
            ref,
            beta0,
            b_vec[1],
            b_vec[2],
            b_vec[3],
            lmu,
        )
    return res_as


@nb.njit(cache=True)
def expanded_qed(ref, order, beta0, b_vec, lmu):
    r"""Compute QED expanded solution at a given order.

    Parameters
    ----------
    ref : float
        reference value of the QED coupling
    order : int
        QED order
    beta0 : float
        first coefficient of the beta function
    b_vec : list
        list of b function coefficients (including b0)
    lmu : float
            logarithm of the ratio between target and reference scales

    Returns
    -------
    float
        QED coupling at target scale :math:`a_em(\mu_R^2)`
    """
    res_aem = ref
    # LO
    if order == 1:
        res_aem = exact_lo(ref, beta0, lmu)
    # NLO
    if order == 2:
        res_aem = expanded_nlo(ref, beta0, b_vec[1], lmu)
    return res_aem


@nb.njit(cache=True)
def couplings_expanded_alphaem_running(
    order, couplings_ref, nf, scale_from, scale_to, decoupled_running
):
    r"""Compute coupled expanded expression of the couplings for running alphaem.

    Implement Eqs. (17-18) from :cite:`Surguladze:1996hx`

    Parameters
    ----------
    order : tuple(int, int)
        perturbation order
    couplings_ref : numpy.ndarray
        reference alpha_s and alpha
    nf : int
        value of nf for computing the couplings
    scale_from : float
        reference scale
    scale_to : float
        target scale
    decoupled_running : bool
        whether the running of the couplings is decoupled or not

    Returns
    -------
    numpy.ndarray
        couplings at target scale :math:`a(\mu_R^2)`
    """
    # common vars
    lmu = np.log(scale_to / scale_from)
    beta0_qcd = beta_qcd((2, 0), nf)
    b_vec_qcd = [b_qcd((i + 2, 0), nf) for i in range(order[0])]
    res_as = expanded_qcd(couplings_ref[0], order[0], beta0_qcd, b_vec_qcd, lmu)
    beta0_qed = beta_qed((0, 2), nf)
    b_vec_qed = [b_qed((0, i + 2), nf) for i in range(order[1])]
    res_aem = expanded_qed(couplings_ref[1], order[1], beta0_qed, b_vec_qed, lmu)
    # if order[0] >= 1 and order[1] >= 1:
    # order[0] is always >=1
    if not decoupled_running:
        if order[1] >= 1:
            res_as += (
                -couplings_ref[0] ** 2
                * b_qcd((2, 1), nf)
                * np.log(1 + beta0_qcd * couplings_ref[1] * lmu)
            )
            res_aem += (
                -couplings_ref[1] ** 2
                * b_qed((1, 2), nf)
                * np.log(1 + beta0_qed * couplings_ref[0] * lmu)
            )
    return np.array([res_as, res_aem])


@nb.njit(cache=True)
def couplings_expanded_fixed_alphaem(order, couplings_ref, nf, scale_from, scale_to):
    r"""Compute coupled expanded expression of the couplings for fixed alphaem.

    Parameters
    ----------
    order : tuple(int, int)
        perturbation order
    couplings_ref : numpy.ndarray
        reference alpha_s and alpha
    nf : int
        value of nf for computing the couplings
    scale_from : float
        reference scale
    scale_to : float
        target scale

    Returns
    -------
    numpy.ndarray
        couplings at target scale :math:`a(\mu_R^2)`
    """
    # common vars
    lmu = np.log(scale_to / scale_from)
    res_as = couplings_ref[0]
    aem = couplings_ref[1]
    beta_qcd0 = beta_qcd((2, 0), nf)
    if order[1] >= 1:
        beta_qcd0 += aem * beta_qcd((2, 1), nf)
    b_vec = [beta_qcd((i + 2, 0), nf) / beta_qcd0 for i in range(order[0])]
    # LO
    if order[0] == 1:
        res_as = exact_lo(couplings_ref[0], beta_qcd0, lmu)
    # NLO
    if order[0] == 2:
        res_as = expanded_nlo(couplings_ref[0], beta_qcd0, b_vec[1], lmu)
    # NNLO
    if order[0] == 3:
        res_as = expanded_nnlo(couplings_ref[0], beta_qcd0, b_vec[1], b_vec[2], lmu)
    # N3LO
    if order[0] == 4:
        res_as = expanded_n3lo(
            couplings_ref[0],
            beta_qcd0,
            b_vec[1],
            b_vec[2],
            b_vec[3],
            lmu,
        )
    return np.array([res_as, aem])


class Couplings:
    r"""Compute the strong and electromagnetic coupling constants :math:`a_s, a_{em}`.

    Note that

    - although, we only provide methods for
      :math:`a_i = \frac{\alpha_i(\mu^2)}{4\pi}` the reference value has to be
      given in terms of :math:`\alpha_i(\mu_0^2)` due to legacy reasons
    - the ``order`` refers to the perturbative order of the beta function, thus
      ``order=(0,0)`` means leading order beta function, means evolution with :math:`\beta_qcd_as2`,
      means running at 1-loop - so there is a natural mismatch between ``order`` and the
      number of loops by one unit

    Normalization is given by :cite:`Herzog:2017ohr`:

    .. math::
        \frac{da_s(\mu^2)}{d\ln\mu^2} = \beta(a_s) \
        = - \sum\limits_{n=0} \beta_n a_s^{n+2}(\mu^2) \quad
        \text{with}~ a_s = \frac{\alpha_s(\mu^2)}{4\pi}

    See :doc:`pQCD ingredients </theory/pQCD>`.

    Parameters
    ----------
    couplings :
        reference configuration
    order :
        Evaluated order of the beta function: ``0`` = LO, ...
    method :
        Applied method to solve the beta function
    masses :
        list with quark masses squared
    hqm_scheme :
        heavy quark mass scheme
    thresholds_ratios :
        list with ratios between the matching scales and the mass squared
    """

    def __init__(
        self,
        couplings: CouplingsInfo,
        order: Order,
        method: CouplingEvolutionMethod,
        masses: List[SquaredScale],
        hqm_scheme: QuarkMassScheme,
        thresholds_ratios: Iterable[float],
    ):
        # Sanity checks
        def assert_positive(name, var):
            if var <= 0:
                raise ValueError(f"{name} has to be positive - got: {var}")

        assert_positive("alpha_s_ref", couplings.alphas)
        assert_positive("alpha_em_ref", couplings.alphaem)
        assert_positive("scale_ref", couplings.scale)
        if order[0] not in [1, 2, 3, 4]:
            raise NotImplementedError("a_s beyond N3LO is not implemented")
        if order[1] not in [0, 1, 2]:
            raise NotImplementedError("a_em beyond NLO is not implemented")
        self.order = tuple(order)
        if method.value not in ["expanded", "exact"]:
            raise ValueError(f"Unknown method {method.value}")
        self.method = method.value

        nf_ref = couplings.num_flavs_ref
        scheme_name = hqm_scheme.name
        self.alphaem_running = couplings.em_running
        self.decoupled_running = False

        # create new threshold object
        self.a_ref = np.array(couplings.values) / 4.0 / np.pi  # convert to a_s and a_em
        matching_scales = (np.array(masses) * np.array(thresholds_ratios)).tolist()
        self.thresholds_ratios = list(thresholds_ratios)
        self.atlas = matchings.Atlas(matching_scales, (couplings.scale**2, nf_ref))
        self.hqm_scheme = scheme_name
        logger.info(
            "Strong Coupling: a_s(µ_R^2=%f)%s=%f=%f/(4π)",
            self.mu2_ref,
            f"^(nf={nf_ref})" if nf_ref else "",
            self.a_ref[0],
            self.a_ref[0] * 4 * np.pi,
        )
        if self.order[1] > 0:
            logger.info(
                "Electromagnetic Coupling: a_em(µ_R^2=%f)%s=%f=%f/(4π)\nalphaem"
                " running: %r\ndecoupled running: %r",
                self.mu2_ref,
                f"^(nf={nf_ref})" if nf_ref else "",
                self.a_ref[1],
                self.a_ref[1] * 4 * np.pi,
                self.alphaem_running,
                self.decoupled_running,
            )
        # cache
        self.cache = {}

    @property
    def mu2_ref(self):
        """Return reference scale."""
        return self.atlas.origin[0]

    def unidimensional_exact(self, beta0, b_vec, u, a_ref, method, rtol):
        """Compute single coupling via decoupled |RGE|.

        Parameters
        ----------
        beta0 : float
            first coefficient of the beta function
        b_vec : list
            list of b function coefficients (including b0)
        u : float
            :math:`log(scale_to / scale_from)`
        a_ref : float
            reference alpha_s or alpha
        method : string
            method for solving the RGE
        rtol : float
            relative acuracy of the solution

        Returns
        -------
        float
            coupling at target scale :math:`a(Q^2)`
        """
        if len(b_vec) == 1:
            return exact_lo(a_ref, beta0, u)

        def rge(_t, a, b_vec):
            rge = -(a**2) * (np.sum([a**k * b for k, b in enumerate(b_vec)]))
            return rge

        res = scipy.integrate.solve_ivp(
            rge,
            (0, beta0 * u),
            (a_ref,),
            args=[b_vec],
            method=method,
            rtol=rtol,
        )
        return res.y[0][-1]

    def compute_exact_alphaem_running(self, a_ref, nf, scale_from, scale_to):
        """Compute couplings via |RGE| with running alphaem.

        Parameters
        ----------
        as_ref : numpy.ndarray
            reference alpha_s and alpha
        nf : int
            value of nf for computing alpha_i
        scale_from : float
            reference scale
        scale_to : float
            target scale

        Returns
        -------
        numpy.ndarray
            couplings at target scale :math:`a(Q^2)`
        """
        # in LO fallback to expanded, as this is the full solution
        u = np.log(scale_to / scale_from)

        if self.order == (1, 0):
            return couplings_expanded_fixed_alphaem(
                self.order, a_ref, nf, scale_from, float(scale_to)
            )

        beta_qcd_vec = [beta_qcd((2, 0), nf)]
        beta_qcd_mix = 0
        beta_qed_mix = 0
        # NLO
        if self.order[0] >= 2:
            beta_qcd_vec.append(beta_qcd((3, 0), nf))
            # NNLO
            if self.order[0] >= 3:
                beta_qcd_vec.append(beta_qcd((4, 0), nf))
                # N3LO
                if self.order[0] >= 4:
                    beta_qcd_vec.append(beta_qcd((5, 0), nf))
        if self.order[1] == 0:
            b_qcd_vec = [
                beta_qcd_vec[i] / beta_qcd_vec[0] for i in range(self.order[0])
            ]
            rge_qcd = self.unidimensional_exact(
                beta_qcd_vec[0],
                b_qcd_vec,
                u,
                a_ref[0],
                method="Radau",
                rtol=1e-6,
            )
            # for order = (qcd, 0) with qcd > 1 we return the exact solution for the QCD RGE
            # while aem is constant
            return np.array([rge_qcd, a_ref[1]])
        if self.order[1] >= 1:
            beta_qed_vec = [beta_qed((0, 2), nf)]
            if not self.decoupled_running:
                beta_qcd_mix = beta_qcd((2, 1), nf)
                beta_qed_mix = beta_qed((1, 2), nf)  # order[0] is always at least 1
            if self.order[1] >= 2:
                beta_qed_vec.append(beta_qed((0, 3), nf))

        # integration kernel
        def rge(_t, a, beta_qcd_vec, beta_qcd_mix, beta_qed_vec, beta_qed_mix):
            rge_qcd = -(a[0] ** 2) * (
                np.sum([a[0] ** k * b for k, b in enumerate(beta_qcd_vec)])
                + a[1] * beta_qcd_mix
            )
            rge_qed = -(a[1] ** 2) * (
                np.sum([a[1] ** k * b for k, b in enumerate(beta_qed_vec)])
                + a[0] * beta_qed_mix
            )
            res = np.array([rge_qcd, rge_qed])
            return res

        # let scipy solve
        res = scipy.integrate.solve_ivp(
            rge,
            (0, u),
            a_ref,
            args=[beta_qcd_vec, beta_qcd_mix, beta_qed_vec, beta_qed_mix],
            method="Radau",
            rtol=1e-6,
        )
        return np.array([res.y[0][-1], res.y[1][-1]])

    def compute_exact_fixed_alphaem(self, a_ref, nf, scale_from, scale_to):
        """Compute couplings via |RGE| with fixed alphaem.

        Parameters
        ----------
        as_ref : numpy.ndarray
            reference alpha_s and alpha
        nf : int
            value of nf for computing alpha_i
        scale_from : float
            reference scale
        scale_to : float
            target scale

        Returns
        -------
        numpy.ndarray
            couplings at target scale :math:`a(Q^2)`
        """
        u = np.log(scale_to / scale_from)

        if self.order in [(1, 0), (1, 1)]:
            return couplings_expanded_fixed_alphaem(
                self.order, a_ref, nf, scale_from, float(scale_to)
            )

        beta_vec = [beta_qcd((2, 0), nf)]
        # NLO
        if self.order[0] >= 2:
            beta_vec.append(beta_qcd((3, 0), nf))
            # NNLO
            if self.order[0] >= 3:
                beta_vec.append(beta_qcd((4, 0), nf))
                # N3LO
                if self.order[0] >= 4:
                    beta_vec.append(beta_qcd((5, 0), nf))
        if self.order[1] >= 1:
            beta_vec[0] += a_ref[1] * beta_qcd((2, 1), nf)
        b_vec = [i / beta_vec[0] for i in beta_vec]
        rge_qcd = self.unidimensional_exact(
            beta_vec[0],
            b_vec,
            u,
            a_ref[0],
            method="Radau",
            rtol=1e-6,
        )
        return np.array([rge_qcd, a_ref[1]])

    def compute(self, a_ref, nf, scale_from, scale_to):
        """Compute actual couplings.

        This is a wrapper in order to pass the computation to the corresponding method
        (depending on the calculation method).

        Parameters
        ----------
        a_ref : numpy.ndarray
            reference a
        nf : int
            value of nf for computing alpha
        scale_from : float
            reference scale
        scale_to : float
            target scale

        Returns
        -------
        numpy.ndarray
            couplings at target scale :math:`a(Q^2)`
        """
        key = (float(a_ref[0]), float(a_ref[1]), nf, scale_from, float(scale_to))
        try:
            return self.cache[key].copy()
        except KeyError:
            # at the moment everything is expanded - and type has been checked in the constructor
            if self.method == "exact":
                if self.alphaem_running:
                    a_new = self.compute_exact_alphaem_running(
                        a_ref.astype(float), nf, scale_from, scale_to
                    )
                else:
                    a_new = self.compute_exact_fixed_alphaem(
                        a_ref.astype(float), nf, scale_from, scale_to
                    )
            else:
                if self.alphaem_running:
                    a_new = couplings_expanded_alphaem_running(
                        self.order,
                        a_ref.astype(float),
                        nf,
                        scale_from,
                        float(scale_to),
                        self.decoupled_running,
                    )
                else:
                    a_new = couplings_expanded_fixed_alphaem(
                        self.order, a_ref.astype(float), nf, scale_from, float(scale_to)
                    )
            self.cache[key] = a_new.copy()
            return a_new

    def a(
        self,
        scale_to,
        nf_to=None,
    ):
        r"""Compute couplings :math:`a_i(\mu_R^2) = \frac{\alpha_i(\mu_R^2)}{4\pi}`.

        Parameters
        ----------
        scale_to : float
            final scale to evolve to :math:`\mu_R^2`
        nf_to : int
            final nf value

        Returns
        -------
        numpy.ndarray
            couplings :math:`a_i(\mu_R^2) = \frac{\alpha_i(\mu_R^2)}{4\pi}`
        """
        # Set up the path to follow in order to go from mu2_0 to mu2_ref
        final_a = self.a_ref.copy()
        path = self.atlas.path((scale_to, nf_to))
        is_downward = matchings.is_downward_path(path)
        shift = matchings.flavor_shift(is_downward)

        for k, seg in enumerate(path):
            # skip a very short segment, but keep the matching
            if not np.isclose(seg.origin, seg.target):
                new_a = self.compute(final_a, seg.nf, seg.origin, seg.target)
            else:
                new_a = final_a
            # apply matching conditions: see hep-ph/9706430
            # - if there is yet a step to go
            if k < len(path) - 1:
                L = np.log(self.thresholds_ratios[seg.nf - shift])
                m_coeffs = (
                    compute_matching_coeffs_down(self.hqm_scheme, seg.nf - 1)
                    if is_downward
                    else compute_matching_coeffs_up(self.hqm_scheme, seg.nf)
                )
                fact = 1.0
                # shift
                for n in range(1, self.order[0]):
                    for l_pow in range(n + 1):
                        fact += new_a[0] ** n * L**l_pow * m_coeffs[n, l_pow]
                new_a[0] *= fact
            final_a = new_a
        return final_a

    def a_s(self, scale_to, nf_to=None):
        r"""Compute strong coupling.

        The strong oupling uses the normalization :math:`a_s(\mu_R^2) =
        \frac{\alpha_s(\mu_R^2)}{4\pi}`.

        Parameters
        ----------
        scale_to : float
            final scale to evolve to :math:`\mu_R^2`
        nf_to : int
            final nf value

        Returns
        -------
        a_s : float
            couplings :math:`a_s(\mu_R^2) = \frac{\alpha_s(\mu_R^2)}{4\pi}`

        """
        return self.a(scale_to, nf_to)[0]

    def a_em(self, scale_to, nf_to=None):
        r"""Compute electromagnetic coupling.

        The electromagnetic oupling uses the normalization :math:`a_em(\mu_R^2)
        = \frac{\alpha_em(\mu_R^2)}{4\pi}`.

        Parameters
        ----------
        scale_to : float
            final scale to evolve to :math:`\mu_R^2`
        nf_to : int
            final nf value

        Returns
        -------
        a_em : float
            couplings :math:`a_em(\mu_R^2) = \frac{\alpha_em(\mu_R^2)}{4\pi}`
        """
        return self.a(scale_to, nf_to)[1]


def compute_matching_coeffs_up(mass_scheme, nf):
    r"""Compute the upward matching coefficients.

    The matching coefficients :cite:`Schroder:2005hy,Chetyrkin:2005ia,Vogt:2004ns`
    are normalized at threshold when moving to a regime with *more* flavors.

    We follow notation of :cite:`Vogt:2004ns` (eq 2.43) for POLE masses.

    The *inverse* |MSbar| values instead are given in :cite:`Schroder:2005hy` (eq 3.1)
    multiplied by a factor of 4 (and 4^2 ...).

    .. math::
        a_s^{(n_l+1)} = a_s^{(n_l)} + \sum\limits_{n=1} (a_s^{(n_l)})^n
                                \sum\limits_{k=0}^n c_{nl} \log(\mu_R^2/\mu_F^2)

    Parameters
    ----------
    mass_scheme : str
        Heavy quark mass scheme: "POLE" or "MSBAR"
    nf : int
        number of active flavors in the lower patch

    Returns
    -------
    numpy.ndarray
        forward matching coefficient matrix
    """
    matching_coeffs_up = np.zeros((4, 4))
    if mass_scheme == "MSBAR":
        matching_coeffs_up[2, 0] = -22.0 / 9.0
        matching_coeffs_up[2, 1] = 22.0 / 3.0

        # c30 = -d30
        matching_coeffs_up[3, 0] = -62.2116 + 5.4177 * nf
        # c31 = -d31 + 5 c11 * c20
        matching_coeffs_up[3, 1] = 365.0 / 3.0 - 67.0 / 9.0 * nf
        matching_coeffs_up[3, 2] = 109.0 / 3.0 + 16.0 / 9.0 * nf

    elif mass_scheme == "POLE":
        matching_coeffs_up[2, 0] = 14.0 / 3.0
        matching_coeffs_up[2, 1] = 38.0 / 3.0
        matching_coeffs_up[3, 0] = 340.729 - 16.7981 * nf
        matching_coeffs_up[3, 1] = 8941.0 / 27.0 - 409.0 / 27.0 * nf
        matching_coeffs_up[3, 2] = 511.0 / 9.0

    matching_coeffs_up[1, 1] = 4.0 / 3.0 * constants.TR
    matching_coeffs_up[2, 2] = 4.0 / 9.0
    matching_coeffs_up[3, 3] = 8.0 / 27.0

    return matching_coeffs_up


# inversion of the matching coefficients
def compute_matching_coeffs_down(mass_scheme, nf):
    """Compute the downward matching coefficients.

    This is the inverse function to :meth:`compute_matching_coeffs_up`.

    Parameters
    ----------
    mass_scheme : str
        Heavy quark mass scheme: "POLE" or "MSBAR"
    nf : int
        number of active flavors in the lower patch

    Returns
    -------
    numpy.ndarray
        downward matching coefficient matrix
    """
    c_up = compute_matching_coeffs_up(mass_scheme, nf)
    return invert_matching_coeffs(c_up)


def invert_matching_coeffs(c_up):
    """Compute the perturbative inverse of the matching conditions.

    They can be obtained e.g. via Mathematica by:

    .. code-block:: Mathematica

        Module[{f, g, l, sol},
            f[a_] := a + Sum[d[n, k]*L^k*a^(1 + n), {n, 3}, {k, 0, n}];
            g[a_] := a + Sum[c[n, k]*L^k*a^(1 + n), {n, 3}, {k, 0, n}] /. {c[1, 0] -> 0};
            l = CoefficientList[Normal@Series[f[g[a]], {a, 0, 5}], {a, L}];
            sol = First@
                Solve[{l[[3]] == 0, l[[4]] == 0, l[[5]] == 0},
                Flatten@Table[d[n, k], {n, 3}, {k, 0, n}]];
            Do[Print@r, {r, sol}];
            Print@Series[f[g[a]] /. sol, {a, 0, 5}];
            Print@Series[g[f[a]] /. sol, {a, 0, 5}];
        ]

    Parameters
    ----------
    c_up : numpy.ndarray
        forward matching coefficient matrix

    Returns
    -------
    numpy.ndarray
        downward matching coefficient matrix
    """
    matching_coeffs_down = np.zeros_like(c_up)
    matching_coeffs_down[1, 1] = -c_up[1, 1]
    matching_coeffs_down[2, 0] = -c_up[2, 0]
    matching_coeffs_down[2, 1] = -c_up[2, 1]
    matching_coeffs_down[2, 2] = 2.0 * c_up[1, 1] ** 2 - c_up[2, 2]
    matching_coeffs_down[3, 0] = -c_up[3, 0]
    matching_coeffs_down[3, 1] = 5 * c_up[1, 1] * c_up[2, 0] - c_up[3, 1]
    matching_coeffs_down[3, 2] = 5 * c_up[1, 1] * c_up[2, 1] - c_up[3, 2]
    matching_coeffs_down[3, 3] = (
        -5 * c_up[1, 1] ** 3 + 5 * c_up[1, 1] * c_up[2, 2] - c_up[3, 3]
    )
    return matching_coeffs_down



"""Define operators container and computing workflow.

The first is the driver class of eko as it is the one that collects all the
previously instantiated information and does the actual computation of the Q2s.

"""
import logging
from dataclasses import astuple
from typing import Dict, List, Optional

import numpy as np
import numpy.typing as npt

from .. import member
from .. import scale_variations as sv
from ..couplings import Couplings
from ..interpolation import InterpolatorDispatcher
from ..io.runcards import Configs, Debug
from ..io.types import EvolutionPoint as EPoint
from ..io.types import Order
from ..matchings import Atlas, Segment, flavor_shift, is_downward_path
from . import Operator, flavors, matching_condition, physical
from .operator_matrix_element import OperatorMatrixElement

logger = logging.getLogger(__name__)

OpDict = Dict[str, Optional[npt.NDArray]]
"""In particular, only the ``operator`` and ``error`` fields are expected."""


class OperatorGrid(sv.ModeMixin):
    """Collection of evolution operators for several scales.

    The operator grid is the driver class of the evolution.

    It receives as input a threshold holder and a generator of a_s.
    From that point onwards it can compute any operator at any q2.

    Attributes
    ----------
    config: dict
    q2_grid: np.ndarray
    managers: dict

    """

    def __init__(
        self,
        mu2grid: List[EPoint],
        order: Order,
        masses: List[float],
        mass_scheme,
        thresholds_ratios: List[float],
        intrinsic_flavors: list,
        xif: float,
        n3lo_ad_variation: tuple,
        configs: Configs,
        debug: Debug,
        atlas: Atlas,
        couplings: Couplings,
        interpol_dispatcher: InterpolatorDispatcher,
    ):
        # check
        config = {}
        config["order"] = order
        config["intrinsic_range"] = intrinsic_flavors
        config["xif2"] = xif**2
        config["HQ"] = mass_scheme
        config["ModSV"] = configs.scvar_method
        config["n3lo_ad_variation"] = n3lo_ad_variation

        for i, q in enumerate("cbt"):
            config[f"m{q}"] = masses[i]
        config["thresholds_ratios"] = thresholds_ratios
        method = config["method"] = configs.evolution_method.value
        config["backward_inversion"] = configs.inversion_method
        config["ev_op_max_order"] = configs.ev_op_max_order
        config["ev_op_iterations"] = configs.ev_op_iterations
        config["n_integration_cores"] = configs.n_integration_cores
        config["debug_skip_singlet"] = debug.skip_singlet
        config["debug_skip_non_singlet"] = debug.skip_non_singlet
        config["polarized"] = configs.polarized
        config["time_like"] = configs.time_like

        if order == (1, 0) and method != "iterate-exact":
            logger.warning("Evolution: In LO we use the exact solution always!")

        logger.info(dict(polarized=configs.polarized))
        logger.info(dict(time_like=configs.time_like))

        self.config = config
        self.q2_grid = mu2grid
        self.managers = dict(
            thresholds_config=atlas,
            couplings=couplings,
            interpol_dispatcher=interpol_dispatcher,
        )
        self._threshold_operators = {}
        self._matching_operators = {}

    def get_threshold_operators(self, path: List[Segment]) -> List[Operator]:
        """Generate the threshold operators.

        This method is called everytime the OperatorGrid is asked for a grid on Q^2
        with a list of the relevant areas.
        If new threshold operators need to be computed, they will be
        cached in an internal dictionary.

        The internal dictionary is self._threshold_operators and its structure is:
        (q2_from, q2_to) -> eko.operators.Operator

        It computes and stores the necessary macthing operators.

        """
        # The base area is always that of the reference q
        thr_ops = []
        # is_downward point to smaller nf
        is_downward = is_downward_path(path)
        shift = flavor_shift(is_downward)
        for seg in path[:-1]:
            new_op_key = astuple(seg)
            kthr = self.config["thresholds_ratios"][seg.nf - shift]
            ome = OperatorMatrixElement(
                self.config,
                self.managers,
                seg.nf - shift + 3,
                seg.target,
                is_downward,
                np.log(kthr),
                self.config["HQ"] == "MSBAR",
            )
            if new_op_key not in self._threshold_operators:
                # Compute the operator and store it
                logger.info("Prepare threshold operator")
                op_th = Operator(self.config, self.managers, seg, is_threshold=True)
                op_th.compute()
                self._threshold_operators[new_op_key] = op_th
            thr_ops.append(self._threshold_operators[new_op_key])

            # Compute the matching conditions and store it
            if seg.target not in self._matching_operators:
                ome.compute()
                self._matching_operators[seg.target] = ome.op_members
        return thr_ops

    def compute(self) -> Dict[EPoint, dict]:
        """Compute all ekos for the `q2grid`."""
        return {q2: self.generate(q2) for q2 in self.q2_grid}

    def generate(self, q2: EPoint) -> OpDict:
        r"""Compute a single EKO.

        eko :math:`\mathbf E(Q^2 \leftarrow Q_0^2)` in flavor basis as numpy array.

        """
        # The lists of areas as produced by the thresholds
        path = self.managers["thresholds_config"].path(q2)
        # Prepare the path for the composition of the operator
        thr_ops = self.get_threshold_operators(path)
        # we start composing with the highest operator ...
        operator = Operator(self.config, self.managers, path[-1])
        operator.compute()

        is_downward = is_downward_path(path)
        intrinsic_range = [4, 5, 6] if is_downward else self.config["intrinsic_range"]
        qed = self.config["order"][1] > 0

        final_op = physical.PhysicalOperator.ad_to_evol_map(
            operator.op_members, operator.nf, operator.q2_to, intrinsic_range, qed
        )
        # and multiply the lower ones from the right
        for op in reversed(list(thr_ops)):
            phys_op = physical.PhysicalOperator.ad_to_evol_map(
                op.op_members, op.nf, op.q2_to, intrinsic_range, qed
            )

            # join with the basis rotation, since matching requires c+ (or likewise)
            nf_match = op.nf - 1 if is_downward else op.nf
            matching = matching_condition.MatchingCondition.split_ad_to_evol_map(
                self._matching_operators[op.q2_to],
                nf_match,
                op.q2_to,
                intrinsic_range=intrinsic_range,
                qed=qed,
            )
            if is_downward:
                invrot = member.ScalarOperator.promote_names(
                    flavors.rotate_matching_inverse(op.nf, qed), op.q2_to
                )
                final_op = final_op @ matching @ invrot @ phys_op
            else:
                rot = member.ScalarOperator.promote_names(
                    flavors.rotate_matching(op.nf + 1, qed), op.q2_to
                )
                final_op = final_op @ rot @ matching @ phys_op
        values, errors = final_op.to_flavor_basis_tensor(qed)
        return {"operator": values, "error": errors}



"""Inventory items definition."""
import io
from dataclasses import asdict, dataclass
from typing import BinaryIO, Optional, Union

import lz4.frame
import numpy as np
import numpy.typing as npt
from numpy.lib import npyio

from .. import matchings
from . import exceptions
from .types import EvolutionPoint as EPoint
from .types import FlavorIndex, FlavorsNumber, SquaredScale


@dataclass(frozen=True)
class Header:
    """Item header, containing metadata."""


@dataclass(frozen=True)
class Evolution(Header):
    """Information to compute an evolution operator.

    It describes the evolution with a fixed number of light flavors between two
    scales.

    """

    origin: SquaredScale
    """Starting point."""
    target: SquaredScale
    """Final point."""
    nf: FlavorsNumber
    """Number of active flavors."""

    cliff: bool = False
    """Whether the operator is reaching a matching scale.

    Cliff operators are the only ones allowed to be intermediate, even though
    they can also be final segments of an evolution path (see
    :meth:`eko.matchings.Atlas.path`).

    Intermediate ones always have final scales :attr:`mu2` corresponding to
    matching scales, and initial scales :attr:`mu20` corresponding to either
    matching scales or the global initial scale of the |EKO|.

    Note
    ----

    The name of *cliff* operators stems from the following diagram::

        nf = 3 --------------------------------------------------------
                        |
        nf = 4 --------------------------------------------------------
                                |
        nf = 5 --------------------------------------------------------
                                                            |
        nf = 6 --------------------------------------------------------

    where each lane corresponds to |DGLAP| evolution with the relative number
    of running flavors, and the vertical bridges are the perturbative matchings
    between two different "adjacent" schemes.

    """

    @classmethod
    def from_atlas(cls, segment: matchings.Segment, cliff: bool = False):
        """Create instance from analogous :class:`eko.matchings.Atlas` object."""
        return cls(**asdict(segment), cliff=cliff)

    @property
    def as_atlas(self) -> matchings.Segment:
        """The associated segment."""
        return matchings.Segment(self.origin, self.target, self.nf)


@dataclass(frozen=True)
class Matching(Header):
    """Information to compute a matching operator.

    Describe the matching between two different flavor number schemes.

    """

    scale: SquaredScale
    hq: FlavorIndex
    inverse: bool = False

    @classmethod
    def from_atlas(cls, matching: matchings.Matching, inverse: bool = False):
        """Create instance from analogous :class:`eko.matchings.Atlas` object."""
        return cls(**asdict(matching), inverse=inverse)

    @property
    def as_atlas(self) -> matchings.Matching:
        """The associated segment."""
        return matchings.Matching(self.scale, self.hq)


Recipe = Union[Evolution, Matching]


@dataclass(frozen=True)
class Target(Header):
    """Target evolution point, labeling evolution from origin to there."""

    scale: SquaredScale
    nf: FlavorsNumber

    @classmethod
    def from_ep(cls, ep: EPoint):
        """Create instance from the :class:`EPoint` analogue."""
        return cls(*ep)

    @property
    def ep(self) -> EPoint:
        """Cast to :class:`EPoint`."""
        return (self.scale, self.nf)


@dataclass(frozen=True)
class Operator:
    """Operator representation.

    To be used to hold the result of a computed 4-dim operator (either a raw
    evolution operator or a matching condition).

    Note
    ----
    IO works with streams in memory, in order to avoid intermediate write on
    disk (keep read from and write to tar file only).

    """

    operator: npt.NDArray
    """Content of the evolution operator."""
    error: Optional[npt.NDArray] = None
    """Errors on individual operator elements (mainly used for integration
    error, but it can host any kind of error).
    """

    def save(self, stream: BinaryIO) -> bool:
        """Save content of operator to bytes.

        The content is saved on a `stream`, in order to be able to perform the
        operation both on disk and in memory.

        The returned value tells whether the operator saved contained or not
        the error (this control even the format, ``npz`` with errors, ``npy``
        otherwise).

        """
        aux = io.BytesIO()
        if self.error is None:
            np.save(aux, self.operator)
        else:
            np.savez(aux, operator=self.operator, error=self.error)
        aux.seek(0)

        # compress if requested
        content = lz4.frame.compress(aux.read())

        # write compressed data
        stream.write(content)
        stream.seek(0)

        # return the type of array dumped (i.e. 'npy' or 'npz')
        return self.error is None

    @classmethod
    def load(cls, stream: BinaryIO):
        """Load operator from bytes.

        An input `stream` is used to load the operator from, in order to
        support the operation both on disk and in memory.

        """
        extracted_stream = io.BytesIO(lz4.frame.decompress(stream.read()))
        content = np.load(extracted_stream)

        if isinstance(content, np.ndarray):
            op = content
            err = None
        elif isinstance(content, npyio.NpzFile):
            op = content["operator"]
            err = content["error"]
        else:
            # TODO: We might consider dropping this exception since np.load will always
            # return a array (or fail on it's own)
            raise exceptions.OperatorLoadingError(
                "Not possible to load operator, content format not recognized"
            )

        return cls(operator=op, error=err)


@dataclass(frozen=True)
class Item:
    """Inventory item."""

    header: Header
    content: Operator



"""Launch EKO calculations, with legacy mu2grid mode."""
import pathlib

import click
import rich

from eko.io import EKO

from .base import command
from .library import OUTPUT


@command.group("inspect")
@click.option(
    "-p",
    "--path",
    type=click.Path(path_type=pathlib.Path, exists=True),
    default=pathlib.Path.cwd() / OUTPUT,
)
@click.pass_context
def subcommand(ctx, path: pathlib.Path):
    """Inspect EKO content."""
    ctx.obj = ctx.with_resource(EKO.read(path))


@subcommand.command("mu2grid")
@click.pass_obj
def sub_mu2(operator: EKO):
    """Check operator's mu2grid."""
    rich.print_json(data=operator.mu2grid.tolist())


@subcommand.command("cards")
@click.pass_obj
def sub_cards(operator: EKO):
    """Check operator's mu2grid."""
    try:
        theory_card = operator.theory_card.raw
    except KeyError as e:
        theory_card = dict(error=f"Key '{e.args}' missing in theory_card")

    try:
        operator_card = operator.operator_card.raw
    except KeyError as e:
        operator_card = dict(error=f"Key '{e.args}' missing in operator_card")

    rich.print_json(data=dict(theory=theory_card, operator=operator_card))



"""Manipulate output generate by EKO."""
import logging
import warnings
from typing import Callable, Optional, Union

import numpy as np
import numpy.typing as npt

from .. import basis_rotation as br
from .. import interpolation
from ..interpolation import XGrid
from .struct import EKO, Operator

logger = logging.getLogger(__name__)

TARGETGRID_ROTATION = "ij,ajbk->aibk"
INPUTGRID_ROTATION = "ajbk,kl->ajbl"
SIMGRID_ROTATION = "ij,ajbk,kl->aibl"
"""Simultaneous grid rotation contraction indices."""

Basis = Union[XGrid, npt.NDArray]


def rotation(new: Optional[Basis], old: Basis, check: Callable, compute: Callable):
    """Define grid rotation.

    This function returns the new grid to be assigned and the rotation computed,
    if the checks for a non-trivial new grid are passed.

    However, the check and the computation are delegated respectively to the
    callables `check` and `compute`.

    """
    if new is None:
        return old, None

    if check(new, old):
        warnings.warn("The new grid is close to the current one")
        return old, None

    return new, compute(new, old)


def xgrid_check(new: Optional[XGrid], old: XGrid):
    """Check validity of new xgrid."""
    return new is not None and len(new) == len(old) and np.allclose(new.raw, old.raw)


def xgrid_compute_rotation(new: XGrid, old: XGrid, interpdeg: int, swap: bool = False):
    """Compute rotation from old to new xgrid.

    By default, the roation is computed for a target xgrid. Whether the function
    should be used for an input xgrid, the `swap` argument should be set to
    `True`, in order to compute it in the other direction (i.e. the transposed).

    """
    if swap:
        new, old = old, new
    b = interpolation.InterpolatorDispatcher(old, interpdeg, False)
    return b.get_interpolation(new.raw)


def xgrid_reshape(
    eko: EKO,
    targetgrid: Optional[XGrid] = None,
    inputgrid: Optional[XGrid] = None,
):
    """Reinterpolate operators on output and/or input grids.

    Target corresponds to the output PDF.

    The operation is in-place.

    """
    eko.assert_permissions(write=True)

    # calling with no arguments is an error
    if targetgrid is None and inputgrid is None:
        raise ValueError("Nor inputgrid nor targetgrid was given")

    interpdeg = eko.operator_card.configs.interpolation_polynomial_degree
    check = xgrid_check
    crot = xgrid_compute_rotation

    # construct matrices
    newtarget, targetrot = rotation(
        targetgrid,
        eko.bases.targetgrid,
        check,
        lambda new, old: crot(new, old, interpdeg),
    )
    newinput, inputrot = rotation(
        inputgrid,
        eko.bases.inputgrid,
        check,
        lambda new, old: crot(new, old, interpdeg, swap=True),
    )

    # after the checks: if there is still nothing to do, skip
    if targetrot is None and inputrot is None:
        logger.debug("Nothing done.")
        return
    # if no rotation is done, the grids are not modified
    if targetrot is not None:
        eko.bases.targetgrid = newtarget
    if inputrot is not None:
        eko.bases.inputgrid = newinput

    # build new grid
    for ep, elem in eko.items():
        assert elem is not None

        operands = [elem.operator]
        operands_errs = [elem.error]

        if targetrot is not None and inputrot is None:
            contraction = TARGETGRID_ROTATION
        elif inputrot is not None and targetrot is None:
            contraction = INPUTGRID_ROTATION
        else:
            contraction = SIMGRID_ROTATION

        if targetrot is not None:
            operands.insert(0, targetrot)
            operands_errs.insert(0, targetrot)
        if inputrot is not None:
            operands.append(inputrot)
            operands_errs.append(inputrot)

        new_operator = np.einsum(contraction, *operands, optimize="optimal")
        if elem.error is not None:
            new_error = np.einsum(contraction, *operands_errs, optimize="optimal")
        else:
            new_error = None

        eko[ep] = Operator(operator=new_operator, error=new_error)

    eko.update()


TARGETPIDS_ROTATION = "ca,ajbk->cjbk"
INPUTPIDS_ROTATION = "ajbk,bd->ajdk"
SIMPIDS_ROTATION = "ca,ajbk,bd->cjdk"
"""Simultaneous grid rotation contraction indices."""


def flavor_reshape(
    eko: EKO,
    targetpids: Optional[npt.NDArray] = None,
    inputpids: Optional[npt.NDArray] = None,
    update: bool = True,
):
    """Change the operators to have in the output targetpids and/or in the input inputpids.

    The operation is in-place.

    Parameters
    ----------
    eko :
        the operator to be rotated
    targetpids :
        target rotation specified in the flavor basis
    inputpids :
        input rotation specified in the flavor basis
    update :
        update :class:`~eko.io.struct.EKO` metadata after writing

    """
    eko.assert_permissions(write=True)

    # calling with no arguments is an error
    if targetpids is None and inputpids is None:
        raise ValueError("Nor inputpids nor targetpids was given")
    # now check to the current status
    if targetpids is not None and np.allclose(
        targetpids, np.eye(len(eko.bases.targetpids))
    ):
        targetpids = None
        warnings.warn("The new targetpids is close to current basis")
    if inputpids is not None and np.allclose(
        inputpids, np.eye(len(eko.bases.inputpids))
    ):
        inputpids = None
        warnings.warn("The new inputpids is close to current basis")
    # after the checks: if there is still nothing to do, skip
    if targetpids is None and inputpids is None:
        logger.debug("Nothing done.")
        return

    # flip input around
    if inputpids is not None:
        inv_inputpids = np.linalg.inv(inputpids)

    # build new grid
    for q2, elem in eko.items():
        ops = elem.operator
        errs = elem.error
        if targetpids is not None and inputpids is None:
            ops = np.einsum(TARGETPIDS_ROTATION, targetpids, ops, optimize="optimal")
            errs = (
                np.einsum(TARGETPIDS_ROTATION, targetpids, errs, optimize="optimal")
                if errs is not None
                else None
            )
        elif inputpids is not None and targetpids is None:
            ops = np.einsum(INPUTPIDS_ROTATION, ops, inv_inputpids, optimize="optimal")
            errs = (
                np.einsum(INPUTPIDS_ROTATION, errs, inv_inputpids, optimize="optimal")
                if errs is not None
                else None
            )
        else:
            ops = np.einsum(
                SIMPIDS_ROTATION, targetpids, ops, inv_inputpids, optimize="optimal"
            )
            errs = (
                np.einsum(
                    SIMPIDS_ROTATION,
                    targetpids,
                    errs,
                    inv_inputpids,
                    optimize="optimal",
                )
                if errs is not None
                else None
            )

        eko[q2] = Operator(operator=ops, error=errs)

    # drop PIDs - keeping them int nevertheless
    # there is no meaningful way to set them in general, after rotation
    if inputpids is not None:
        eko.bases.inputpids = np.array([0] * len(eko.bases.inputpids))
    if targetpids is not None:
        eko.bases.targetpids = np.array([0] * len(eko.bases.targetpids))

    if update:
        eko.update()


def to_evol(eko: EKO, source: bool = True, target: bool = False):
    """Rotate the operator into evolution basis.

    This assigns also the pids. The operation is in-place.

    Parameters
    ----------
    eko :
        the operator to be rotated
    source :
        rotate on the input tensor
    target :
        rotate on the output tensor

    """
    # rotate
    inputpids = br.rotate_flavor_to_evolution if source else None
    targetpids = br.rotate_flavor_to_evolution if target else None
    # prevent metadata update, since flavor_reshape has not enough information
    # to determine inpupids and targetpids, and they will be updated after the
    # call
    flavor_reshape(eko, inputpids=inputpids, targetpids=targetpids, update=False)
    # assign pids
    if source:
        eko.bases.inputpids = inputpids
    if target:
        eko.bases.targetpids = targetpids

    eko.update()


def to_uni_evol(eko: EKO, source: bool = True, target: bool = False):
    """Rotate the operator into evolution basis.

    This assigns also the pids. The operation is in-place.

    Parameters
    ----------
    eko :
        the operator to be rotated
    source :
        rotate on the input tensor
    target :
        rotate on the output tensor

    """
    # rotate
    inputpids = br.rotate_flavor_to_unified_evolution if source else None
    targetpids = br.rotate_flavor_to_unified_evolution if target else None
    # prevent metadata update, since flavor_reshape has not enough information
    # to determine inpupids and targetpids, and they will be updated after the
    # call
    flavor_reshape(eko, inputpids=inputpids, targetpids=targetpids, update=False)
    # assign pids
    if source:
        eko.bases.inputpids = inputpids
    if target:
        eko.bases.targetpids = targetpids

    eko.update()



r"""Holds the classes that define the |FNS|."""
import logging
from dataclasses import dataclass
from typing import List, Union

import numpy as np

from .io.types import EvolutionPoint as EPoint
from .io.types import FlavorIndex, FlavorsNumber, SquaredScale
from .quantities.heavy_quarks import MatchingScales

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Segment:
    """Oriented path in the threshold landscape."""

    origin: SquaredScale
    """Starting point."""
    target: SquaredScale
    """Final point."""
    nf: FlavorsNumber
    """Number of active flavors."""

    @property
    def is_downward(self) -> bool:
        """Return True if ``origin`` is bigger than ``target``."""
        return self.origin > self.target

    def __str__(self):
        """Textual representation, mainly for logging purpose."""
        return f"Segment({self.origin} -> {self.target}, nf={self.nf})"


Path = List[Segment]


@dataclass(frozen=True)
class Matching:
    """Matching between two different segments.

    The meaning of the flavor index `hq` is the PID of the corresponding heavy
    quark.

    """

    scale: SquaredScale
    hq: FlavorIndex


MatchedPath = List[Union[Segment, Matching]]


class Atlas:
    r"""Holds information about the matching scales.

    These scales are the :math:`Q^2` has to pass in order to get there from a
    given :math:`Q^2_{ref}`.

    """

    def __init__(self, matching_scales: MatchingScales, origin: EPoint):
        """Create basic atlas."""
        self.walls = [0] + matching_scales + [np.inf]
        self.origin = self.normalize(origin)

        logger.info(str(self))

    def normalize(self, target: EPoint) -> EPoint:
        """Fill number of flavors if needed."""
        if target[1] is not None:
            return target
        return (target[0], nf_default(target[0], self))

    def __str__(self):
        """Textual representation, mainly for logging purpose."""
        walls = " - ".join([f"{w:.2e}" for w in self.walls])
        return f"Atlas [{walls}], ref={self.origin[0]} @ {self.origin[1]}"

    @classmethod
    def ffns(cls, nf: int, mu2: SquaredScale):
        """Create a |FFNS| setup.

        The function creates simply sufficient thresholds at ``0`` (in the
        beginning), since the number of flavors is determined by counting
        from below.

        The origin is set with that number of flavors.

        """
        matching_scales = MatchingScales([0] * (nf - 3) + [np.inf] * (6 - nf))
        origin = (mu2, nf)
        return cls(matching_scales, origin)

    def path(self, target: EPoint) -> Path:
        """Determine the path to the target evolution point.

        Essentially, the path is always monotonic in the number of flavors,
        increasing or decreasing the active flavors by one unit every time a
        matching happens at the suitable scale.

        Examples
        --------
        Since this can result in a counter-intuitive behavior, let's walk through some examples.

        Starting with the intuitive one:
        >>> Atlas([10, 20, 30], (5, 3)).path((25, 5))
        [Segment(5, 10, 3), Segment(10, 20, 4), Segment(20, 25, 5)]

        If the number of flavor has been reached, it will continue walking
        without matchin again.
        >>> Atlas([10, 20, 30], (5, 3)).path((25, 4))
        [Segment(5, 10, 3), Segment(10, 25, 4)]

        It is irrelevant the scale you start from, to step from 3 to 4 you have
        to cross the charm matching scale, whether this means walking upward or
        downward.
        >>> Atlas([10, 20, 30], (15, 3)).path((25, 5))
        [Segment(15, 10, 3), Segment(10, 20, 4), Segment(20, 25, 5)]

        An actual backward evolution is defined by lowering the number of
        flavors going from origin to target.
        >>> Atlas([10, 20, 30], (25, 5)).path((5, 3))
        [Segment(25, 20, 5), Segment(20, 10, 4), Segment(10, 5, 3)]

        But the only difference is in the matching between two segments, since
        a single segment is always happening in a fixed number of flavors, and
        it is completely analogue for upward or downward evolution.

        Note
        ----

        Since the only task required to determine a path is interleaving the
        correct matching scales, this is done by slicing the walls.

        """
        mu20, nf0 = self.origin
        mu2f, nff = self.normalize(target)

        # determine direction and python slice modifier
        rc, shift = (-1, -3) if nff < nf0 else (1, -2)

        # join all necessary points in one list
        boundaries = [mu20] + self.walls[nf0 + shift : nff + shift : rc] + [mu2f]

        return [
            Segment(boundaries[i], mu2, nf0 + i * rc)
            for i, mu2 in enumerate(boundaries[1:])
        ]

    def matched_path(self, target: EPoint) -> MatchedPath:
        """Determine the path to the target, including matchings.

        In practice, just a wrapper around :meth:`path` adding the intermediate
        matchings.

        """
        path = self.path(target)

        prev = path[0]
        matched: MatchedPath = [prev]
        for seg in path[1:]:
            matching = Matching(prev.target, max(prev.nf, seg.nf))

            matched.append(matching)
            matched.append(seg)
            prev = seg

        return matched


def nf_default(mu2: SquaredScale, atlas: Atlas) -> FlavorsNumber:
    r"""Determine the number of active flavors in the *default flow*.

    Default flow is defined by the natural sorting of the matching scales:

    .. math::

        \mu_c < \mu_b < \mu_t

    So, the flow is defined starting with 3 flavors below the charm matching,
    and increasing by one every time a matching scale is passed while
    increasing the scale.

    """
    ref_idx = np.digitize(mu2, atlas.walls)
    return int(2 + ref_idx)


def is_downward_path(path: Path) -> bool:
    r"""Determine if a path is downward.

    Criterias are:

    - in the number of active flavors when the path list contains more than one
      :class:`Segment`, note this can be different from each
      :attr:`Segment.is_downward`
    - in :math:`\mu^2`, when just one single :class:`Segment` is given

    """
    if len(path) == 1:
        return path[0].is_downward
    return path[1].nf < path[0].nf


def flavor_shift(is_downward: bool) -> int:
    """Determine the shift to number of light flavors."""
    return 4 if is_downward else 3



"""Heavy quarks related quantities."""
import enum
from dataclasses import dataclass
from typing import Generic, List, Sequence, TypeVar

import numpy as np

from ..io.dictlike import DictLike
from ..io.types import (
    FlavorsNumber,
    IntrinsicFlavors,
    LinearScale,
    ReferenceRunning,
    SquaredScale,
)

FLAVORS = "cbt"

T = TypeVar("T")


class HeavyQuarks(list, Generic[T]):
    """Access heavy quarks attributes by name."""

    def __init__(self, args: Sequence[T]):
        if len(args) != 3:
            raise ValueError("Pass values for exactly three quarks.")

        self.extend(args)

    @property
    def c(self) -> T:
        """Charm quark."""
        return self[0]

    @c.setter
    def c(self, value: T):
        self[0] = value

    @property
    def b(self) -> T:
        """Bottom quark."""
        return self[1]

    @b.setter
    def b(self, value: T):
        self[1] = value

    @property
    def t(self) -> T:
        """Top quark."""
        return self[2]

    @t.setter
    def t(self, value: T):
        self[2] = value


QuarkMass = LinearScale
QuarkMassRef = ReferenceRunning[QuarkMass]
HeavyQuarkMasses = HeavyQuarks[QuarkMassRef]
MatchingRatio = float
MatchingRatios = HeavyQuarks[MatchingRatio]
MatchingScale = SquaredScale
MatchingScales = HeavyQuarks[MatchingScale]


# TODO: upgrade the following to StrEnum (requires py>=3.11) with that, it is
# possible to replace all non-alias right sides with calls to enum.auto()


class QuarkMassScheme(enum.Enum):
    """Scheme to define heavy quark masses."""

    MSBAR = "msbar"
    POLE = "pole"


@dataclass
class HeavyInfo(DictLike):
    """Collect information about heavy quarks.

    This is meant to be used mainly as a theory card section, and to be passed
    around when all or a large part of this information is required.

    Note
    ----
    All scales and ratios in this structure are linear, so you can consider
    them as quantities in :math:`GeV` or ratios of them.

    """

    num_flavs_init: FlavorsNumber
    r"""Number of active flavors at fitting scale.

    I.e. :math:`n_{f,\text{ref}}(\mu^2_0)`, formerly called ``nf0``.

    """
    num_flavs_max_pdf: FlavorsNumber
    """Maximum number of quark PDFs."""
    intrinsic_flavors: IntrinsicFlavors
    """List of intrinsic quark PDFs."""
    masses: HeavyQuarkMasses
    """List of heavy quark masses."""
    masses_scheme: QuarkMassScheme
    """Scheme used to specify heavy quark masses."""
    matching_ratios: MatchingRatios
    """Matching scale of heavy quark masses"""

    @property
    def squared_ratios(self) -> List[float]:
        """Squared ratios of matching scales."""
        return np.power(self.matching_ratios, 2.0).tolist()



"""LHAPDF info file utilities."""
import copy
import math

import numpy as np

from eko import couplings
from eko.io.runcards import OperatorCard, TheoryCard

from .genpdf import load


def build(
    theory_card: TheoryCard,
    operators_card: OperatorCard,
    num_members: int,
    info_update: dict,
):
    """Generate a lhapdf info file from theory and operators card.

    Parameters
    ----------
    theory_card : dict
        theory card
    operators_card : dict
        operators_card
    num_members : int
        number of pdf set members
    info_update : dict
        info to update

    Returns
    -------
    dict
        info file in lhapdf format
    """
    template_info = copy.deepcopy(load.template_info)
    template_info["SetDesc"] = "Evolved PDF from " + str(operators_card.mu0) + " GeV"
    template_info["Authors"] = ""
    template_info["FlavorScheme"] = "variable"
    template_info.update(info_update)
    template_info["NumFlavors"] = 14
    template_info["Flavors"] = [-6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 21, 22]
    # TODO actually point to input grid
    template_info["XMin"] = float(operators_card.xgrid.raw[0])
    template_info["XMax"] = float(operators_card.xgrid.raw[-1])
    template_info["NumMembers"] = num_members
    template_info["OrderQCD"] = theory_card.order[0] - 1
    template_info["QMin"] = round(math.sqrt(operators_card.mu2grid[0]), 4)
    template_info["QMax"] = round(math.sqrt(operators_card.mu2grid[-1]), 4)
    template_info["MZ"] = theory_card.couplings.scale
    template_info["MUp"] = 0.0
    template_info["MDown"] = 0.0
    template_info["MStrange"] = 0.0
    template_info["MCharm"] = theory_card.heavy.masses.c.value
    template_info["MBottom"] = theory_card.heavy.masses.b.value
    template_info["MTop"] = theory_card.heavy.masses.t.value
    template_info["AlphaS_MZ"] = theory_card.couplings.alphas
    template_info["AlphaS_OrderQCD"] = theory_card.order[0] - 1
    evmod = couplings.couplings_mod_ev(operators_card.configs.evolution_method)
    quark_masses = [(x.value) ** 2 for x in theory_card.heavy.masses]
    sc = couplings.Couplings(
        theory_card.couplings,
        theory_card.order,
        evmod,
        quark_masses,
        hqm_scheme=theory_card.heavy.masses_scheme,
        thresholds_ratios=np.power(list(iter(theory_card.heavy.matching_ratios)), 2.0),
    )
    alphas_values = np.array(
        [
            4.0
            * np.pi
            * sc.a_s(
                muf2,
            )
            for muf2 in operators_card.mu2grid
        ],
        dtype=float,
    )
    template_info["AlphaS_Vals"] = alphas_values.tolist()
    template_info["AlphaS_Qs"] = np.array(
        [mu for mu, _ in operators_card.mugrid]
    ).tolist()
    return template_info



"""Tools to generate runcards."""
import os
from math import nan

import numpy as np
import yaml

from eko.io import runcards
from eko.io.types import RawCard, ReferenceRunning

_theory = dict(
    order=[1, 0],
    couplings=dict(
        alphas=0.118,
        alphaem=0.007496252,
        scale=91.2,
        num_flavs_ref=5,
        max_num_flavs=6,
    ),
    heavy=dict(
        num_flavs_init=4,
        num_flavs_max_pdf=6,
        intrinsic_flavors=[4],
        masses=[ReferenceRunning([mq, nan]) for mq in (2.0, 4.5, 173.07)],
        masses_scheme="POLE",
        matching_ratios=[1.0, 1.0, 1.0],
    ),
    xif=1.0,
    n3lo_ad_variation=(0, 0, 0, 0),
)

_operator = dict(
    mu0=1.65,
    mugrid=[(100.0, 5)],
    xgrid=np.geomspace(1e-7, 1.0, 50).tolist(),
    configs=dict(
        evolution_method="iterate-exact",
        ev_op_max_order=[10, 0],
        ev_op_iterations=10,
        interpolation_polynomial_degree=4,
        interpolation_is_log=True,
        scvar_method=None,
        inversion_method=None,
        n_integration_cores=0,
        polarized=False,
        time_like=False,
    ),
    debug=dict(
        skip_singlet=False,
        skip_non_singlet=False,
    ),
)


class example:
    """Provide runcards examples."""

    @classmethod
    def theory(cls) -> runcards.TheoryCard:
        """Provide example theory card object."""
        return runcards.TheoryCard.from_dict(_theory)

    @classmethod
    def operator(cls) -> runcards.OperatorCard:
        """Provide example operator card object."""
        return runcards.OperatorCard.from_dict(_operator)

    @classmethod
    def raw_theory(cls):
        """Provide example theory card unstructured."""
        # TODO: consider to return instead `cls.theory().raw`
        return _theory.copy()

    @classmethod
    def raw_operator(cls):
        """Provide example operator card unstructured."""
        return _operator.copy()


def dump(card: RawCard, path: os.PathLike):
    """Export the operators card.

    Parameters
    ----------
    card : dict
        card to dump
    path : str
        destination of the dumped card

    """
    with open(path, "w", encoding="utf-8") as fd:
        yaml.safe_dump(card, fd)


def load(path) -> RawCard:
    """Import the theory card specified by path.

    Parameters
    ----------
    path : str
        path to theory card in yaml format

    Returns
    -------
    dict
        loaded card

    """
    with open(path, encoding="utf-8") as fd:
        card = yaml.safe_load(fd)

    return card



"""LHAPDF interface."""
import numpy as np

from eko import basis_rotation as br


def compute_LHAPDF_data(
    theory, operators, pdf, skip_pdfs, rotate_to_evolution_basis=False
):
    """
    Run LHAPDF to compute operators.

    Parameters
    ----------
    operators : dict
        operators card
    pdf : lhapdf_type
        pdf
    skip_pdfs : list
        list of pdfs (pid or name) to skip
    rotate_to_evolution_basis: bool
        rotate to evolution basis

    Returns
    -------
    ref : dict
        output containing: target_xgrid, values

    """
    target_xgrid = operators["interpolation_xgrid"]

    out_tabs = {}
    for mu2 in np.array(operators["mugrid"]) ** 2:
        tab = {}
        for pid in br.flavor_basis_pids:
            if pid in skip_pdfs:
                continue

            # collect lhapdf
            me = []
            for x in target_xgrid:
                xf = pdf.xfxQ2(pid, x, mu2)
                me.append(xf)
            tab[pid] = np.array(me)

        # rotate if needed
        if rotate_to_evolution_basis:
            qed = theory["QED"] > 0
            if not qed:
                evol_basis = br.evol_basis
                rotate_flavor_to_evolution = br.rotate_flavor_to_evolution
            else:
                evol_basis = br.unified_evol_basis
                rotate_flavor_to_evolution = br.rotate_flavor_to_unified_evolution
            pdfs = np.array(
                [
                    tab[pid] if pid in tab else np.zeros(len(target_xgrid))
                    for pid in br.flavor_basis_pids
                ]
            )
            evol_pdf = rotate_flavor_to_evolution @ pdfs
            tab = dict(zip(evol_basis, evol_pdf))

        out_tabs[mu2] = tab

    ref = {
        "target_xgrid": target_xgrid,
        "values": out_tabs,
    }

    return ref



"""Tools to evolve actual PDFs."""
import pathlib

import numpy as np

import eko
from eko import basis_rotation as br
from eko.io import EKO

from . import apply, genpdf, info_file

DEFAULT_NAME = "eko.tar"


def evolve_pdfs(
    initial_PDF_list,
    theory_card,
    operators_card,
    path=None,
    store_path=None,
    targetgrid=None,
    install=False,
    name="Evolved_PDF",
    info_update=None,
):
    """Evolves one or several initial_PDFs and dump the evolved PDFs in lhapdf format.

    Parameters
    ----------
    initial_PDF_list : list(lhapdf object)
        list of PDF members to be evolved
    theory_card : dict
        theory card
    operators_card : dict
        operators card
    path : str
        path to cached eko output (if "None" it will be recomputed)
    store_path : str
        path where the eko is stored (if "None" will not be saved)
    targetgrid : list(float)
        target x-grid (if different from input x-grid)
    install : bool
        set whether to install evolved PDF to lhapdf directory
    name : str
        set name of evolved PDF
    info_update : dict
        dict of info to add or update to default info file
    """
    # update op and th cards
    if path is not None:
        eko_path = pathlib.Path(path)
        if eko_path.is_dir():
            eko_path = eko_path / DEFAULT_NAME
    else:
        if store_path is None:
            raise ValueError("'store_path' required if 'path' is not provided.")
        eko.solve(theory_card, operators_card, path=store_path)
        eko_path = store_path

    evolved_PDF_list = []
    with EKO.read(eko_path) as eko_output:
        for initial_PDF in initial_PDF_list:
            evolved_PDF_list.append(
                apply.apply_pdf(eko_output, initial_PDF, targetgrid)
            )

    if targetgrid is None:
        targetgrid = operators_card.xgrid
    if info_update is None:
        info_update = {}
    info_update["XMin"] = targetgrid.raw[0]
    info_update["XMax"] = targetgrid.raw[-1]
    info = info_file.build(
        theory_card,
        operators_card,
        len(evolved_PDF_list),
        info_update=info_update,
    )
    all_member_blocks = []
    targetlist = targetgrid.raw.tolist()
    for evolved_PDF in evolved_PDF_list:
        all_blocks = []
        block = genpdf.generate_block(
            lambda pid, x, Q2, evolved_PDF=evolved_PDF: targetlist[targetlist.index(x)]
            * evolved_PDF[Q2]["pdfs"][pid][targetlist.index(x)],
            xgrid=targetlist,
            evolgrid=operators_card.evolgrid,
            pids=np.array(br.flavor_basis_pids),
        )
        # all_blocks will be useful in case there will be necessity to dump many blocks
        # for a single member
        all_blocks.append(block)
        all_member_blocks.append(all_blocks)

    genpdf.export.dump_set(name, info, all_member_blocks)

    if install:
        genpdf.install_pdf(name)
